"""Collection of Mesh Functions"""

# global
import ivy
from functools import reduce
from operator import mul

# local
from ivy_vision import single_view_geometry as ivy_svg

MIN_DENOMINATOR = 1e-12


def rasterize_triangles(
    pixel_coords_triangles, image_dims, batch_shape=None, device=None
):
    """Rasterize image-projected triangles based on:
    https://www.scratchapixel.com/lessons/3d-basic-rendering/rasterization-practical
    -implementation/rasterization-stage and:
    https://www.scratchapixel.com/lessons/3d-basic-rendering/rasterization-practical
    -implementation/rasterization-practical-implementation

    Parameters
    ----------
    pixel_coords_triangles
        Projected image-space triangles to be rasterized
        *[batch_shape,input_size,3,3]*
    image_dims
        Image dimensions.
    batch_shape
        Shape of batch. Inferred from Inputs if None. (Default value = None)
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        Rasterized triangles

    """
    if batch_shape is None:
        batch_shape = []

    if device is None:
        device = ivy.dev(pixel_coords_triangles)

    # shapes as list
    batch_shape = list(batch_shape)
    num_batch_dims = len(batch_shape)
    image_dims = list(image_dims)
    input_image_dims = pixel_coords_triangles.shape[num_batch_dims:-2]
    input_image_dims_prod = reduce(mul, input_image_dims, 1)

    # BS x 3 x 2
    pixel_xy_coords = pixel_coords_triangles[..., 0:2]

    # BS x 3 x 1
    pixel_x_coords = pixel_coords_triangles[..., 0:1]
    pixel_y_coords = pixel_coords_triangles[..., 1:2]

    # 1
    x_min = ivy.reshape(ivy.min(pixel_x_coords, keepdims=True), (-1,))
    x_max = ivy.reshape(ivy.max(pixel_x_coords, keepdims=True), (-1,))
    x_range = x_max - x_min
    y_min = ivy.reshape(ivy.min(pixel_y_coords, keepdims=True), (-1,))
    y_max = ivy.reshape(ivy.max(pixel_y_coords, keepdims=True), (-1,))
    y_range = y_max - y_min

    # 2
    bbox = ivy.concat((x_range, y_range), axis=0)
    img_bbox_list = [
        int(item)
        for item in ivy.to_list(ivy.concat((y_range + 1, x_range + 1), axis=0))
    ]

    # BS x 2
    v0 = pixel_xy_coords[..., 0, :]
    v1 = pixel_xy_coords[..., 1, :]
    v2 = pixel_xy_coords[..., 2, :]
    tri_centres = (v0 + v1 + v2) / 3

    # BS x 1
    v0x = v0[..., 0:1]
    v0y = v0[..., 1:2]
    v1x = v1[..., 0:1]
    v1y = v1[..., 1:2]
    v2x = v2[..., 0:1]
    v2y = v2[..., 1:2]

    # BS x BBX x BBY x 2
    uniform_sample_coords = ivy_svg.create_uniform_pixel_coords_image(
        img_bbox_list, batch_shape
    )[..., 0:2]
    P = ivy.round(uniform_sample_coords + tri_centres - bbox / 2)

    # BS x BBX x BBY x 1
    Px = P[..., 0:1]
    Py = P[..., 1:2]
    v0v1_edge_func = ((Px - v0x) * (v1y - v0y) - (Py - v0y) * (v1x - v0x)) >= 0
    v1v2_edge_func = ((Px - v1x) * (v2y - v1y) - (Py - v1y) * (v2x - v1x)) >= 0
    v2v0_edge_func = ((Px - v2x) * (v0y - v2y) - (Py - v2y) * (v0x - v2x)) >= 0
    edge_func = ivy.logical_and(
        ivy.logical_and(v0v1_edge_func, v1v2_edge_func), v2v0_edge_func
    )

    batch_indices_list = list()
    for i, batch_dim in enumerate(batch_shape):
        # get batch shape
        batch_dims_before = batch_shape[:i]
        num_batch_dims_before = len(batch_dims_before)
        batch_dims_after = batch_shape[i + 1 :]
        num_batch_dims_after = len(batch_dims_after)

        # [batch_dim]
        batch_indices = ivy.arange(batch_dim, dtype="int32", device=device)

        # [1]*num_batch_dims_before x batch_dim x [1]*num_batch_dims_after x 1 x 1
        reshaped_batch_indices = ivy.reshape(
            batch_indices,
            [1] * num_batch_dims_before
            + [batch_dim]
            + [1] * num_batch_dims_after
            + [1, 1],
        )

        # BS x N x 1
        tiled_batch_indices = ivy.tile(
            reshaped_batch_indices,
            batch_dims_before + [1] + batch_dims_after + [input_image_dims_prod * 9, 1],
        )
        batch_indices_list.append(tiled_batch_indices)

    # BS x N x (num_batch_dims + 2)
    all_indices = ivy.concat(
        batch_indices_list
        + [
            ivy.astype(
                ivy.flip(ivy.reshape(P, batch_shape + [-1, 2]), axis=-1), "int32"
            )
        ],
        axis=-1,
    )

    # offset uniform images
    return ivy.astype(
        ivy.flip(
            ivy.scatter_nd(
                ivy.reshape(all_indices, [-1, num_batch_dims + 2]),
                ivy.reshape(ivy.astype(edge_func, "int32"), (-1, 1)),
                shape=batch_shape + image_dims + [1],
                reduction="sum",
            ),
            axis=-3,
        ),
        "bool",
    )


def create_trimesh_indices_for_image(batch_shape, image_dims, device=None):
    """Create triangle mesh for image with given image dimensions

    Parameters
    ----------
    batch_shape
        Shape of batch.
    image_dims
        Image dimensions.
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        (Default value = None)

    Returns
    -------
    ret
        Triangle mesh indices for image *[batch_shape,h*w*some_other_stuff,3]*

    """
    # shapes as lists
    batch_shape = list(batch_shape)
    image_dims = list(image_dims)

    # other shape specs
    num_batch_dims = len(batch_shape)
    tri_dim = 2 * (image_dims[0] - 1) * (image_dims[1] - 1)
    flat_shape = [1] * num_batch_dims + [tri_dim] + [3]
    tile_shape = batch_shape + [1] * 2

    # 1 x W-1
    t00_ = ivy.reshape(
        ivy.arange(image_dims[1] - 1, dtype="float32", device=device), (1, -1)
    )

    # H-1 x 1
    k_ = (
        ivy.reshape(
            ivy.arange(image_dims[0] - 1, dtype="float32", device=device), (-1, 1)
        )
        * image_dims[1]
    )

    # H-1 x W-1
    t00_ = ivy.matmul(ivy.ones((image_dims[0] - 1, 1), device=device), t00_)
    k_ = ivy.matmul(k_, ivy.ones((1, image_dims[1] - 1), device=device))

    # (H-1xW-1) x 1
    t00 = ivy.expand_dims(t00_ + k_, axis=-1)
    t01 = t00 + 1
    t02 = t00 + image_dims[1]
    t10 = t00 + image_dims[1] + 1
    t11 = t01
    t12 = t02

    # (H-1xW-1) x 3
    t0 = ivy.concat((t00, t01, t02), axis=-1)
    t1 = ivy.concat((t10, t11, t12), axis=-1)

    # BS x 2x(H-1xW-1) x 3
    return ivy.tile(ivy.reshape(ivy.concat((t0, t1), axis=0), flat_shape), tile_shape)


def coord_image_to_trimesh(
    coord_img, validity_mask=None, batch_shape=None, image_dims=None, device=None
):
    """Create trimesh, with vertices and triangle indices, from co-ordinate image.

    Parameters
    ----------
    coord_img
        Image of co-ordinates *[batch_shape,h,w,3]*
    validity_mask
        Boolean mask of where the coord image contains valid values
        *[batch_shape,h,w,1]* (Default value = None)
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_dims
        Image dimensions. Inferred from inputs in None. (Default value = None)
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        Vertices *[batch_shape,(hxw),3]* amd Trimesh indices *[batch_shape,n,3]*

    """
    if device is None:
        device = ivy.dev(coord_img)

    if batch_shape is None:
        batch_shape = ivy.shape(coord_img)[:-3]

    if image_dims is None:
        image_dims = ivy.shape(coord_img)[-3:-1]

    # shapes as lists
    batch_shape = list(batch_shape)
    image_dims = list(image_dims)

    # BS x (HxW) x 3
    vertices = ivy.reshape(coord_img, batch_shape + [image_dims[0] * image_dims[1], 3])

    if validity_mask is not None:
        # BS x H-1 x W-1 x 1
        t00_validity = validity_mask[
            ..., 0 : image_dims[0] - 1, 0 : image_dims[1] - 1, :
        ]
        t01_validity = validity_mask[..., 0 : image_dims[0] - 1, 1 : image_dims[1], :]
        t02_validity = validity_mask[..., 1 : image_dims[0], 0 : image_dims[1] - 1, :]
        t10_validity = validity_mask[..., 1 : image_dims[0], 1 : image_dims[1], :]
        t11_validity = t01_validity
        t12_validity = t02_validity

        # BS x H-1 x W-1 x 1
        t0_validity = ivy.logical_and(
            t00_validity, ivy.logical_and(t01_validity, t02_validity)
        )
        t1_validity = ivy.logical_and(
            t10_validity, ivy.logical_and(t11_validity, t12_validity)
        )

        # BS x (H-1xW-1)
        t0_validity_flat = ivy.reshape(t0_validity, batch_shape + [-1])
        t1_validity_flat = ivy.reshape(t1_validity, batch_shape + [-1])

        # BS x 2x(H-1xW-1)
        trimesh_index_validity = ivy.concat(
            (t0_validity_flat, t1_validity_flat), axis=-1
        )

        # BS x N
        trimesh_valid_indices = ivy.argwhere(trimesh_index_validity)

        # BS x 2x(H-1xW-1) x 3
        all_trimesh_indices = create_trimesh_indices_for_image(
            batch_shape, image_dims, device
        )

        # BS x N x 3
        trimesh_indices = ivy.gather_nd(all_trimesh_indices, trimesh_valid_indices)

    else:
        # BS x N=2x(H-1xW-1) x 3
        trimesh_indices = create_trimesh_indices_for_image(batch_shape, image_dims)

    # BS x (HxW) x 3,    BS x N x 3
    return vertices, trimesh_indices



"""Collection of Quantization Functions"""

# global
import ivy

# local
from ivy_vision import single_view_geometry as ivy_svg

MIN_DENOMINATOR = 1e-12
MIN_DEPTH_DIFF = 1e-2


def quantize_to_image(
    pixel_coords,
    final_image_dims,
    feat=None,
    feat_prior=None,
    with_db=False,
    pixel_coords_var=1e-3,
    feat_var=1e-3,
    pixel_coords_prior_var=1e12,
    feat_prior_var=1e12,
    var_threshold=(1e-3, 1e12),
    uniform_pixel_coords=None,
    batch_shape=None,
    device=None,
):
    r"""Quantize pixel co-ordinates with d feature channels (for depth, rgb, normals
    etc.), from images :math:`\mathbf{X}\in\mathbb{R}^{input\_images\_shape×(2+d)}`,
    which may have been reprojected from a host of different cameras (leading to
    non-integer pixel values), to a new quantized pixel co-ordinate image with the
    same feature channels :math:`\mathbf{X}\in\mathbb{R}^{h×w×(2+d)}`, and with
    integer pixel co-ordinates. Duplicates during the quantization are either
    probabilistically fused based on variance, or the minimum depth is chosen when
    using depth buffer mode.

    Parameters
    ----------
    pixel_coords
        Pixel co-ordinates *[batch_shape,input_size,2]*
    final_image_dims
        Image dimensions of the final image.
    feat
        Features (i.e. depth, rgb, encoded), default is None.
        *[batch_shape,input_size,d]*
    feat_prior
        Prior feature image mean, default is None. *[batch_shape,input_size,d]*
    with_db
        Whether or not to use depth buffer in rendering, default is false
    pixel_coords_var
        Pixel coords variance *[batch_shape,input_size,2]* (Default value = 1e-3)
    feat_var
        Feature variance *[batch_shape,input_size,d]* (Default value = 1e-3)
    pixel_coords_prior_var
        Pixel coords prior variance *[batch_shape,h,w,2]* (Default value = 1e12)
    feat_prior_var
        Features prior variance *[batch_shape,h,w,3]* (Default value = 1e12)
    var_threshold
        Variance threshold, for projecting valid coords and clipping
        *[batch_shape,2+d,2]* (Default value = (1e-3)
    uniform_pixel_coords
        Homogeneous uniform (integer) pixel co-ordinate images,
        inferred from final_image_dims
        if None *[batch_shape,h,w,3]* (Default value = None)
    batch_shape
        Shape of batch. Assumed no batches if None. (Default value = None)
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        Quantized pixel co-ordinates image with d feature channels
        (for depth, rgb, normals etc.) *[batch_shape,h,w,2+d]*,
        maybe the quantized variance, *[batch_shape,h,w,2+d]*, and scatter counter image
        *[batch_shape,h,w,1]*

    """
    # ToDo: make variance fully optional. If not specified,
    #  then do not compute and scatter during function call for better efficiency.
    # config
    if batch_shape is None:
        batch_shape = pixel_coords.shape[:-2]

    if device is None:
        device = ivy.dev(pixel_coords)

    if feat is None:
        d = 0
    else:
        d = feat.shape[-1]
    min_depth_diff = ivy.array([MIN_DEPTH_DIFF], device=device)
    red = "min" if with_db else "sum"

    # shapes as list
    batch_shape = list(batch_shape)
    final_image_dims = list(final_image_dims)
    num_batch_dims = len(batch_shape)

    # variance threshold
    if isinstance(var_threshold, tuple) or isinstance(var_threshold, list):
        ones = ivy.ones(batch_shape + [1, 2 + d, 1])
        var_threshold = ivy.concat(
            (ones * var_threshold[0], ones * var_threshold[1]), axis=-1
        )
    else:
        var_threshold = ivy.reshape(var_threshold, batch_shape + [1, 2 + d, 2])

    # uniform pixel coords
    if uniform_pixel_coords is None:
        uniform_pixel_coords = ivy_svg.create_uniform_pixel_coords_image(
            final_image_dims, batch_shape, device=device
        )
    uniform_pixel_coords = uniform_pixel_coords[..., 0:2]

    # Extract Values #

    feat_prior = (
        ivy.ones_like(feat) * feat_prior
        if isinstance(feat_prior, float)
        else feat_prior
    )
    pixel_coords_var = (
        ivy.ones_like(pixel_coords) * pixel_coords_var
        if isinstance(pixel_coords_var, float)
        else pixel_coords_var
    )
    feat_var = (
        ivy.ones_like(feat) * feat_var if isinstance(feat_var, float) else feat_var
    )
    pixel_coords_prior_var = (
        ivy.ones(batch_shape + final_image_dims + [2]) * pixel_coords_prior_var
        if isinstance(pixel_coords_prior_var, float)
        else pixel_coords_prior_var
    )
    feat_prior_var = (
        ivy.ones(batch_shape + final_image_dims + [d]) * feat_prior_var
        if isinstance(feat_prior_var, float)
        else feat_prior_var
    )

    # Quantize #

    # BS x N x 2
    quantized_pixel_coords = ivy.reshape(
        ivy.astype(ivy.round(pixel_coords), "int32"), batch_shape + [-1, 2]
    )

    # Combine #

    # BS x N x (2+D)
    pc_n_feat = ivy.reshape(
        ivy.concat((pixel_coords, feat), axis=-1), batch_shape + [-1, 2 + d]
    )
    pc_n_feat_var = ivy.reshape(
        ivy.concat((pixel_coords_var, feat_var), axis=-1), batch_shape + [-1, 2 + d]
    )

    # BS x H x W x (2+D)
    prior = ivy.concat((uniform_pixel_coords, feat_prior), axis=-1)
    prior_var = ivy.concat((pixel_coords_prior_var, feat_prior_var), axis=-1)

    # Validity Mask #

    # BS x N x 1
    var_validity_mask = (
        ivy.sum(
            ivy.astype(pc_n_feat_var < var_threshold[..., 1], "int32"),
            axis=-1,
            keepdims=True,
        )
        == 2 + d
    )
    bounds_validity_mask = ivy.logical_and(
        ivy.logical_and(
            quantized_pixel_coords[..., 0:1] >= 0, quantized_pixel_coords[..., 1:2] >= 0
        ),
        ivy.logical_and(
            quantized_pixel_coords[..., 0:1] <= final_image_dims[1] - 1,
            quantized_pixel_coords[..., 1:2] <= final_image_dims[0] - 1,
        ),
    )
    validity_mask = ivy.logical_and(var_validity_mask, bounds_validity_mask)

    # num_valid_indices x len(BS)+2
    validity_indices = ivy.reshape(
        ivy.astype(ivy.argwhere(validity_mask), "int32"), [-1, num_batch_dims + 2]
    )
    num_valid_indices = validity_indices.shape[-2]

    if num_valid_indices == 0:
        return (
            ivy.concat((uniform_pixel_coords[..., 0:2], feat_prior), axis=-1),
            ivy.concat((pixel_coords_prior_var, feat_prior_var), axis=-1),
            ivy.zeros_like(feat[..., 0:1], device=device),
        )

    # Depth Based Scaling #

    mean_depth_min = None
    mean_depth_range = None
    pc_n_feat_wo_depth_range = None
    pc_n_feat_wo_depth_min = None
    var_vals_range = None
    var_vals_min = None

    if with_db:
        # BS x N x 1
        mean_depth = pc_n_feat[..., 2:3]

        # BS x 1 x 1
        mean_depth_min = ivy.min(mean_depth, axis=-2, keepdims=True)
        mean_depth_max = ivy.max(mean_depth, axis=-2, keepdims=True)
        mean_depth_range = mean_depth_max - mean_depth_min

        # BS x N x 1
        scaled_depth = (mean_depth - mean_depth_min) / (
            mean_depth_range * min_depth_diff + MIN_DENOMINATOR
        )

        if d == 1:
            # BS x 1 x 1+D
            pc_n_feat_wo_depth_min = ivy.zeros(batch_shape + [1, 0], device=device)
            pc_n_feat_wo_depth_range = ivy.ones(batch_shape + [1, 0], device=device)

        else:
            # feat without depth

            # BS x N x 1+D
            pc_n_feat_wo_depth = ivy.concat(
                (pc_n_feat[..., 0:2], pc_n_feat[..., 3:]), axis=-1
            )

            # find the min and max of each value

            # BS x 1 x 1+D
            pc_n_feat_wo_depth_max = (
                ivy.max(pc_n_feat_wo_depth, axis=-2, keepdims=True) + 1
            )
            pc_n_feat_wo_depth_min = (
                ivy.min(pc_n_feat_wo_depth, axis=-2, keepdims=True) - 1
            )
            pc_n_feat_wo_depth_range = pc_n_feat_wo_depth_max - pc_n_feat_wo_depth_min

            # BS x N x 1+D
            normed_pc_n_feat_wo_depth = (
                pc_n_feat_wo_depth - pc_n_feat_wo_depth_min
            ) / (pc_n_feat_wo_depth_range + MIN_DENOMINATOR)

            # combine with scaled depth

            # BS x N x 1+D
            pc_n_feat_wo_depth_scaled = normed_pc_n_feat_wo_depth + scaled_depth

            # BS x N x (2+D)
            pc_n_feat = ivy.concat(
                (
                    pc_n_feat_wo_depth_scaled[..., 0:2],
                    mean_depth,
                    pc_n_feat_wo_depth_scaled[..., 2:],
                ),
                axis=-1,
            )

        # scale variance

        # BS x 1 x (2+D)
        var_vals_max = ivy.max(pc_n_feat_var, axis=-2, keepdims=True) + 1
        var_vals_min = ivy.min(pc_n_feat_var, axis=-2, keepdims=True) - 1
        var_vals_range = var_vals_max - var_vals_min

        # BS x N x (2+D)
        normed_var_vals = (pc_n_feat_var - var_vals_min) / (
            var_vals_range + MIN_DENOMINATOR
        )
        pc_n_feat_var = normed_var_vals + scaled_depth

        # ready for later reversal with full image dimensions

        # BS x 1 x 1 x D
        var_vals_min = ivy.expand_dims(var_vals_min, axis=-2)
        var_vals_range = ivy.expand_dims(var_vals_range, axis=-2)

    # Validity Pruning #

    # num_valid_indices x (2+D)
    pc_n_feat = ivy.gather_nd(pc_n_feat, validity_indices[..., 0 : num_batch_dims + 1])
    pc_n_feat_var = ivy.gather_nd(
        pc_n_feat_var, validity_indices[..., 0 : num_batch_dims + 1]
    )

    # num_valid_indices x 2
    quantized_pixel_coords = ivy.gather_nd(
        quantized_pixel_coords, validity_indices[..., 0 : num_batch_dims + 1]
    )

    if with_db:
        means_to_scatter = pc_n_feat
        vars_to_scatter = pc_n_feat_var
    else:
        # num_valid_indices x (2+D)
        vars_to_scatter = 1 / (pc_n_feat_var + MIN_DENOMINATOR)
        means_to_scatter = pc_n_feat * vars_to_scatter

    # Scatter #

    # num_valid_indices x 1
    counter = ivy.ones_like(pc_n_feat[..., 0:1], device=device)
    if with_db:
        counter *= -1

    # num_valid_indices x 2(2+D)+1
    values_to_scatter = ivy.concat(
        (means_to_scatter, vars_to_scatter, counter), axis=-1
    )

    # num_valid_indices x (num_batch_dims + 2)
    all_indices = ivy.flip(quantized_pixel_coords, axis=-1)
    if num_batch_dims > 0:
        all_indices = ivy.concat((validity_indices[..., :-2], all_indices), axis=-1)

    # BS x H x W x (2(2+D) + 1)
    quantized_img = ivy.scatter_nd(
        ivy.reshape(all_indices, [-1, num_batch_dims + 2]),
        ivy.reshape(values_to_scatter, [-1, 2 * (2 + d) + 1]),
        shape=batch_shape + final_image_dims + [2 * (2 + d) + 1],
        reduction=red,
    )

    # BS x H x W x 1
    quantized_counter = quantized_img[..., -1:]
    if with_db:
        invalidity_mask = quantized_counter != -1
    else:
        invalidity_mask = quantized_counter == 0

    if with_db:
        # BS x H x W x (2+D)
        quantized_mean_scaled = quantized_img[..., 0 : 2 + d]
        quantized_var_scaled = quantized_img[..., 2 + d : 2 * (2 + d)]

        # BS x H x W x 1
        quantized_depth_mean = quantized_mean_scaled[..., 2:3]

        # BS x 1 x 1 x 1
        mean_depth_min = ivy.expand_dims(mean_depth_min, axis=-2)
        mean_depth_range = ivy.expand_dims(mean_depth_range, axis=-2)

        # BS x 1 x 1 x (1+D)
        pc_n_feat_wo_depth_min = ivy.expand_dims(pc_n_feat_wo_depth_min, axis=-2)
        pc_n_feat_wo_depth_range = ivy.expand_dims(pc_n_feat_wo_depth_range, axis=-2)

        # BS x 1 x 1 x (2+D) x 2
        var_threshold = ivy.expand_dims(var_threshold, axis=-3)

        # BS x H x W x (1+D)
        quantized_mean_wo_depth_scaled = ivy.concat(
            (quantized_mean_scaled[..., 0:2], quantized_mean_scaled[..., 3:]), axis=-1
        )
        quantized_mean_wo_depth_normed = quantized_mean_wo_depth_scaled - (
            quantized_depth_mean - mean_depth_min
        ) / (mean_depth_range * min_depth_diff + MIN_DENOMINATOR)
        quantized_mean_wo_depth = (
            quantized_mean_wo_depth_normed * pc_n_feat_wo_depth_range
            + pc_n_feat_wo_depth_min
        )
        prior_wo_depth = ivy.concat((prior[..., 0:2], prior[..., 3:]), axis=-1)
        quantized_mean_wo_depth = ivy.where(
            invalidity_mask, prior_wo_depth, quantized_mean_wo_depth
        )

        # BS x H x W x (2+D)
        quantized_mean = ivy.concat(
            (
                quantized_mean_wo_depth[..., 0:2],
                quantized_depth_mean,
                quantized_mean_wo_depth[..., 2:],
            ),
            axis=-1,
        )

        # BS x H x W x (2+D)
        quantized_var_normed = quantized_var_scaled - (
            quantized_depth_mean - mean_depth_min
        ) / (mean_depth_range * min_depth_diff + MIN_DENOMINATOR)
        quantized_var = ivy.maximum(
            quantized_var_normed * var_vals_range + var_vals_min, var_threshold[..., 0]
        )
        quantized_var = ivy.where(invalidity_mask, prior_var, quantized_var)
    else:
        # BS x H x W x (2+D)
        quantized_sum_mean_x_recip_var = quantized_img[..., 0 : 2 + d]
        quantized_var_wo_increase = ivy.where(
            invalidity_mask,
            prior_var,
            (1 / (quantized_img[..., 2 + d : 2 * (2 + d)] + MIN_DENOMINATOR)),
        )
        quantized_var = ivy.maximum(
            quantized_var_wo_increase * quantized_counter,
            ivy.expand_dims(var_threshold[..., 0], axis=-2),
        )
        quantized_var = ivy.where(invalidity_mask, prior_var, quantized_var)
        quantized_mean = ivy.where(
            invalidity_mask,
            prior,
            quantized_var_wo_increase * quantized_sum_mean_x_recip_var,
        )

    # BS x H x W x (2+D)    BS x H x W x (2+D)     BS x H x W x 1
    return quantized_mean, quantized_var, quantized_counter



from . import containers
from .containers import *
from . import implicit
from .implicit import *
from . import mesh
from .mesh import *
from . import optical_flow
from .optical_flow import *
from . import padding
from .padding import *
from . import projective_geometry
from .projective_geometry import *
from . import quantization
from .quantization import *
from . import sdf
from .sdf import *
from . import single_view_geometry
from .single_view_geometry import *
from . import smoothing
from .smoothing import *
from . import two_view_geometry
from .two_view_geometry import *
from . import voxel_grids
from .voxel_grids import *



"""Collection of Single-View-Geometry Functions"""

# global
import ivy
import numpy as np
import ivy_mech
from operator import mul
from functools import reduce

# local
from ivy_vision import projective_geometry as ivy_pg
from ivy_vision import image as ivy_img
from ivy_vision.containers import Intrinsics
from ivy_vision.containers import Extrinsics
from ivy_vision.containers import CameraGeometry


MIN_DENOMINATOR = 1e-12


def create_uniform_pixel_coords_image(
    image_dims, batch_shape=None, normalized=False, homogeneous=True, device=None
):
    r"""Create image of homogeneous integer
    :math:`xy` pixel co-ordinates
    :math:`\mathbf{X}\in\mathbb{Z}^{h×w×3}`, stored
    as floating point values.
     The origin is at the top-left corner of the image,
     with :math:`+x` rightwards, and
    :math:`+y` downwards. The final homogeneous dimension are all ones.
    In subsequent use of this image, the depth of
    each pixel can be represented using this same homogeneous
    representation, by simply scaling each 3-vector by the
    depth value.
    The final dimension therefore always holds the depth value,
    while the former two dimensions hold depth
    scaled pixel co-ordinates.\n
    `[reference] <localhost:63342/ivy/docs/source/references/mvg_textbook.pdf#page=172>`_ # noqa
    deduction from top of page 154, section 6.1, equation 6.1

    Parameters
    ----------
    image_dims
        Image dimensions.
    batch_shape
        Shape of batch. Assumed no batch dimensions if None.
        (Default value = None)
    normalized
        Whether to normalize x-y pixel co-ordinates to the range 0-1.
        Default is False.
    homogeneous
        Whether the pixel co-ordinates should be 3D homogeneous or just 2D.
        Default is True.
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        (Default value = None)

    Returns
    -------
    ret
        Image of homogeneous pixel co-ordinates *[batch_shape,height,width,3]*

    """
    # shapes as lists
    batch_shape = [] if batch_shape is None else batch_shape
    batch_shape = list(batch_shape)
    image_dims = list(image_dims)

    # other shape specs
    num_batch_dims = len(batch_shape)
    flat_shape = [1] * num_batch_dims + image_dims + [2]
    tile_shape = batch_shape + [1] * 3

    # H x W x 1
    pixel_x_coords = ivy.astype(
        ivy.reshape(
            ivy.tile(ivy.arange(image_dims[1], device=device), [image_dims[0]]),
            (image_dims[0], image_dims[1], 1),
        ),
        "float32",
    )
    if normalized:
        pixel_x_coords = pixel_x_coords / (float(image_dims[1]) + MIN_DENOMINATOR)

    # W x H x 1
    pixel_y_coords_ = ivy.astype(
        ivy.reshape(
            ivy.tile(ivy.arange(image_dims[0], device=device), [image_dims[1]]),
            (image_dims[1], image_dims[0], 1),
        ),
        "float32",
    )

    # H x W x 1
    pixel_y_coords = ivy.permute_dims(pixel_y_coords_, axes=(1, 0, 2))
    if normalized:
        pixel_y_coords = pixel_y_coords / (float(image_dims[0]) + MIN_DENOMINATOR)

    # BS x H x W x 2
    pix_coords = ivy.tile(
        ivy.reshape(ivy.concat((pixel_x_coords, pixel_y_coords), axis=-1), flat_shape),
        tile_shape,
    )

    if homogeneous:
        # BS x H x W x 3
        pix_coords = ivy_mech.make_coordinates_homogeneous(pix_coords)

    # BS x H x W x 2or3
    return pix_coords


def persp_angles_to_focal_lengths(persp_angles, image_dims, device=None):
    r"""Compute focal lengths :math:`f_x, f_y`
    from perspective angles :math:`θ_x, θ_y`.\n
    `[reference] <localhost:63342/ivy/docs/source/references/mvg_textbook.pdf#page=172>`_ # noqa
    deduction from page 154, section 6.1, figure 6.1

    Parameters
    ----------
    persp_angles
        Perspective angles *[batch_shape,2]*
    image_dims
        Image dimensions.
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        Focal lengths *[batch_shape,2]*

    """
    if device is None:
        device = ivy.dev(persp_angles)

    # shapes as list
    image_dims = list(image_dims)

    # BS x 2
    return -ivy.flip(
        ivy.astype(ivy.array(image_dims, device=device), "float32"), axis=-1
    ) / (2 * ivy.tan(persp_angles / 2) + MIN_DENOMINATOR)


def focal_lengths_to_persp_angles(focal_lengths, image_dims, device=None):
    r"""Compute perspective angles
    :math:`θ_x, θ_y` from focal lengths :math:`f_x, f_y`.\n
    `[reference] <localhost:63342/ivy/docs/source/references/mvg_textbook.pdf#page=172>`_ # noqa
    deduction from page 154, section 6.1, figure 6.1

    Parameters
    ----------
    focal_lengths
        batch_shape,2]*
    image_dims
        Image dimensions.
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        Perspective angles *[batch_shape,2]*

    """
    if device is None:
        device = ivy.dev(focal_lengths)

    # shapes as list
    image_dims = list(image_dims)

    # BS x 2
    return -2 * ivy.atan(
        ivy.flip(ivy.astype(ivy.array(image_dims, device=device), "float32"), axis=-1)
        / (2 * focal_lengths + MIN_DENOMINATOR)
    )


def focal_lengths_and_pp_offsets_to_calib_mat(
    focal_lengths, pp_offsets, batch_shape=None, device=None
):
    r"""Compute calibration matrix :math:`\mathbf{K}\in\mathbb{R}^{3×3}`
    from focal lengths :math:`f_x, f_y` and
    principal-point offsets :math:`p_x, p_y`.\n
    `[reference] <localhost:63342/ivy/docs/source/references/mvg_textbook.pdf#page=173>`_ # noqa
    page 155, section 6.1, equation 6.4

    Parameters
    ----------
    focal_lengths
        Focal lengths *[batch_shape,2]*
    pp_offsets
        Principal-point offsets *[batch_shape,2]*
    batch_shape
        Shape of batch. Inferred from inputs if None.
        (Default value = None)
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        Calibration matrix *[batch_shape,3,3]*

    """
    if batch_shape is None:
        batch_shape = focal_lengths.shape[:-1]

    if device is None:
        device = ivy.dev(focal_lengths)

    # shapes as list
    batch_shape = list(batch_shape)

    # BS x 1 x 1
    zeros = ivy.zeros(batch_shape + [1, 1], device=device)
    ones = ivy.ones(batch_shape + [1, 1], device=device)

    # BS x 2 x 1
    focal_lengths_reshaped = ivy.expand_dims(focal_lengths, axis=-1)
    pp_offsets_reshaped = ivy.expand_dims(pp_offsets, axis=-1)

    # BS x 1 x 3
    row1 = ivy.concat(
        (focal_lengths_reshaped[..., 0:1, :], zeros, pp_offsets_reshaped[..., 0:1, :]),
        axis=-1,
    )
    row2 = ivy.concat(
        (zeros, focal_lengths_reshaped[..., 1:2, :], pp_offsets_reshaped[..., 1:2, :]),
        axis=-1,
    )
    row3 = ivy.concat((zeros, zeros, ones), axis=-1)

    # BS x 3 x 3
    return ivy.concat((row1, row2, row3), axis=-2)


# noinspection PyUnresolvedReferences
def rot_mat_and_cam_center_to_ext_mat(rotation_mat, camera_center, batch_shape=None):
    r"""Get extrinsic matrix :math:`\mathbf{E}\in\mathbb{R}^{3×4}`
    from rotation matrix
    :math:`\mathbf{R}\in\mathbb{R}^{3×3}` and camera centers
    :math:`\overset{\sim}{\mathbf{C}}\in\mathbb{R}^{3×1}`.\n
    `[reference] <localhost:63342/ivy/docs/source/references/mvg_textbook.pdf#page=175>`_ # noqa
    page 157, section 6.1, equation 6.11

    Parameters
    ----------
    rotation_mat
        Rotation matrix *[batch_shape,3,3]*
    camera_center
        Camera center *[batch_shape,3,1]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)

    Returns
    -------
    ret
        Extrinsic matrix *[batch_shape,3,4]*

    """
    if batch_shape is None:
        batch_shape = rotation_mat.shape[:-2]

    # shapes as list
    batch_shape = list(batch_shape)

    # num batch dims
    num_batch_dims = len(batch_shape)

    # BS x 3 x 3
    identity = ivy.tile(
        ivy.reshape(ivy.eye(3), [1] * num_batch_dims + [3, 3]), batch_shape + [1, 1]
    )

    # BS x 3 x 4
    identity_w_cam_center = ivy.concat((identity, -camera_center), axis=-1)

    # BS x 3 x 4
    return ivy.matmul(rotation_mat, identity_w_cam_center)


def depth_to_ds_pixel_coords(
    depth, uniform_pixel_coords=None, batch_shape=None, image_dims=None
):
    r"""Get depth scaled homogeneous pixel co-ordinates image
    :math:`\mathbf{X}_p\in\mathbb{R}^{h×w×3}` from depth image
    :math:`\mathbf{X}_d\in\mathbb{R}^{h×w×1}`.\n

    Parameters
    ----------
    depth
        Depth image *[batch_shape,h,w,1]*
    uniform_pixel_coords
        Image of homogeneous pixel co-ordinates. Created if None.
        *[batch_shape,h,w,3]* (Default value = None)
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_dims
        Image dimensions. Inferred from inputs in None. (Default value = None)

    Returns
    -------
    ret
        Depth scaled homogeneous pixel co-ordinates image *[batch_shape,h,w,3]*

    """
    if batch_shape is None:
        batch_shape = depth.shape[:-3]

    if image_dims is None:
        image_dims = depth.shape[-3:-1]

    # shapes as list
    batch_shape = list(batch_shape)
    image_dims = list(image_dims)

    # BS x H x W x 3
    if uniform_pixel_coords is None:
        uniform_pixel_coords = create_uniform_pixel_coords_image(
            image_dims, batch_shape, device=ivy.dev(depth)
        )

    # BS x H x W x 3
    return uniform_pixel_coords * depth


def depth_to_radial_depth(
    depth, inv_calib_mat, uniform_pix_coords=None, batch_shape=None, image_dims=None
):
    r"""Get radial depth image
    :math:`\mathbf{X}_{rd}\in\mathbb{R}^{hxw×1}`
    from depth image
    :math:`\mathbf{X}_d\in\mathbb{R}^{hxw×1}`.\n

    Parameters
    ----------
    depth
        Depth scaled homogeneous pixel co-ordinates image
        *[batch_shape,image_dims,1]*
    inv_calib_mat
        Inverse calibration matrix *[batch_shape,3,3]*
    uniform_pix_coords
        Uniform homogeneous pixel co-ordinates, constructed if None.
        *[batch_shape,image_dims,3]* (Default value = None)
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_dims
        Image shape. Inferred from inputs in None. (Default value = None)

    Returns
    -------
    ret
        Radial depth image *[batch_shape,image_dims,1]*

    """
    if batch_shape is None:
        batch_shape = inv_calib_mat.shape[:-2]
    num_batch_dims = len(batch_shape)

    if image_dims is None:
        image_dims = depth.shape[num_batch_dims:-1]

    # shapes as list
    batch_shape = list(batch_shape)
    image_dims = list(image_dims)

    # BS x H x W x 3
    if uniform_pix_coords is None:
        uniform_pix_coords = create_uniform_pixel_coords_image(
            image_dims, batch_shape, device=ivy.dev(depth)
        )
    ds_pixel_coords = uniform_pix_coords * depth

    # BS x H x W x 1
    return ds_pixel_coords_to_radial_depth(
        ds_pixel_coords, inv_calib_mat, batch_shape, image_dims
    )


def ds_pixel_coords_to_radial_depth(
    ds_pixel_coords, inv_calib_mat, batch_shape=None, image_shape=None
):
    r"""Get radial depth image :math:`\mathbf{X}_{rd}\in\mathbb{R}^{is×1}`
    from depth scaled homogeneous pixel
    co-ordinates image :math:`\mathbf{X}_p\in\mathbb{R}^{is×3}`.\n

    Parameters
    ----------
    ds_pixel_coords
        Depth scaled homogeneous pixel co-ordinates image *[batch_shape,image_shape,1]*
    inv_calib_mat
        Inverse calibration matrix *[batch_shape,3,3]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image shape. Inferred from inputs in None. (Default value = None)

    Returns
    -------
    ret
        Radial depth image *[batch_shape,image_shape,1]*

    """
    if batch_shape is None:
        batch_shape = inv_calib_mat.shape[:-2]
    num_batch_dims = len(batch_shape)

    if image_shape is None:
        image_shape = ds_pixel_coords.shape[num_batch_dims:-1]

    # shapes as list
    batch_shape = list(batch_shape)
    image_shape = list(image_shape)

    # BS x IS x 3
    cam_coords = ds_pixel_to_cam_coords(
        ds_pixel_coords, inv_calib_mat, batch_shape, image_shape
    )[..., 0:3]

    # BS x IS x 1
    return ivy.sum(cam_coords**2, axis=-1, keepdims=True) ** 0.5


def cam_to_ds_pixel_coords(
    coords_wrt_cam, calib_mat, batch_shape=None, image_shape=None
):
    r"""Get depth scaled homogeneous pixel co-ordinates image
    :math:`\mathbf{X}_p\in\mathbb{R}^{is×3}` from camera-centric
    homogeneous co-ordinates image
    :math:`\mathbf{X}_c\in\mathbb{R}^{is×4}`.\n
    `[reference] <localhost:63342/ivy/docs/source/references/mvg_textbook.pdf#page=173>`_ # noqa
    page 155, equation 6.3

    Parameters
    ----------
    coords_wrt_cam
        Camera-centric homogeneous co-ordinates image *[batch_shape,image_shape,4]*
    calib_mat
        Calibration matrix *[batch_shape,3,3]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image shape. Inferred from inputs in None. (Default value = None)

    Returns
    -------
    ret
        Depth scaled homogeneous pixel co-ordinates image *[batch_shape,image_shape,3]*

    """
    if batch_shape is None:
        batch_shape = calib_mat.shape[:-2]
    num_batch_dims = len(batch_shape)

    if image_shape is None:
        image_shape = coords_wrt_cam.shape[num_batch_dims:-1]

    # shapes as list
    batch_shape = list(batch_shape)
    image_shape = list(image_shape)

    # BS x IS x 3
    coords_wrt_cam = coords_wrt_cam[..., 0:3]

    # BS x IS x 3
    return ivy_pg.transform(coords_wrt_cam, calib_mat, batch_shape, image_shape)


def cam_coords_to_depth(coords_wrt_cam, calib_mat, batch_shape=None, image_shape=None):
    r"""Get depth image :math:`\mathbf{X}_p\in\mathbb{R}^{is×1}` from camera-centric
    homogeneous co-ordinates image :math:`\mathbf{X}_c\in\mathbb{R}^{is×4}`.\n

    Parameters
    ----------
    coords_wrt_cam
        Camera-centric homogeneous co-ordinates image *[batch_shape,image_shape,4]*
    calib_mat
        Calibration matrix *[batch_shape,3,3]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image shape. Inferred from inputs in None. (Default value = None)

    Returns
    -------
    ret
        Depth image *[batch_shape,image_shape,1]*

    """
    # BS x IS x 1
    return cam_to_ds_pixel_coords(coords_wrt_cam, calib_mat, batch_shape, image_shape)[
        ..., -1:
    ]


def ds_pixel_to_cam_coords(
    ds_pixel_coords, inv_calib_mat, batch_shape=None, image_shape=None, device=None
):
    r"""Get camera-centric homogeneous co-ordinates image
    :math:`\mathbf{X}_c\in\mathbb{R}^{is×4}` from
    depth scaled homogeneous pixel co-ordinates image
    :math:`\mathbf{X}_p\in\mathbb{R}^{is×3}`.\n
    `[reference] <localhost:63342/ivy/docs/source/references/mvg_textbook.pdf#page=173>`_ # noqa
    page 155, matrix inverse of equation 6.3

    Parameters
    ----------
    ds_pixel_coords
        Depth scaled homogeneous pixel co-ordinates image
        *[batch_shape,image_shap,3]*
    inv_calib_mat
        Inverse calibration matrix *[batch_shape,3,3]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image dimensions. Inferred from inputs in None. (Default value = None)
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        Camera-centric homogeneous co-ordinates image *[batch_shape,image_shape,4]*

    """
    if batch_shape is None:
        batch_shape = inv_calib_mat.shape[:-2]
    num_batch_dims = len(batch_shape)

    if image_shape is None:
        image_shape = ds_pixel_coords.shape[num_batch_dims:-1]

    if device is None:
        device = ivy.dev(ds_pixel_coords)

    # shapes as list
    batch_shape = list(batch_shape)
    image_shape = list(image_shape)

    # BS x IS x 3
    cam_coords = ivy_pg.transform(
        ds_pixel_coords, inv_calib_mat, batch_shape, image_shape
    )

    # BS x IS x 4
    return ivy_mech.make_coordinates_homogeneous(cam_coords, batch_shape + image_shape)


def depth_to_cam_coords(
    depth,
    inv_calib_mat,
    uniform_pixel_coords=None,
    batch_shape=None,
    image_dims=None,
    device=None,
):
    r"""Get camera-centric homogeneous co-ordinates image
    :math:`\mathbf{X}_c\in\mathbb{R}^{hxw×4}` from
    depth image :math:`\mathbf{X}_p\in\mathbb{R}^{hxw×1}`.\n

    Parameters
    ----------
    depth
        Depth image *[batch_shape,h,w,1]*
    inv_calib_mat
        Inverse calibration matrix *[batch_shape,3,3]*
    uniform_pixel_coords
        Image of homogeneous pixel co-ordinates. Created if None.
        *[batch_shape,h,w,3]* (Default value = None)
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_dims
        Image dimensions. Inferred from inputs in None. (Default value = None)
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        Camera-centric homogeneous co-ordinates image *[batch_shape,h,w,4]*

    """
    if batch_shape is None:
        batch_shape = inv_calib_mat.shape[:-2]

    if image_dims is None:
        image_dims = depth.shape[-3:-1]

    if device is None:
        device = ivy.dev(depth)

    # shapes as list
    batch_shape = list(batch_shape)
    image_dims = list(image_dims)

    # BS x H x W x 3
    ds_pixel_coords = depth_to_ds_pixel_coords(
        depth, uniform_pixel_coords, batch_shape, image_dims
    )

    # BS x H x W x 4
    return ds_pixel_to_cam_coords(
        ds_pixel_coords, inv_calib_mat, batch_shape, image_dims, device
    )


def world_to_cam_coords(
    coords_wrt_world, ext_mat, batch_shape=None, image_shape=None, device=None
):
    r"""Get camera-centric homogeneous co-ordinates image
    :math:`\mathbf{X}_c\in\mathbb{R}^{is×4}` from world-centric
    homogeneous co-ordinates image
    :math:`\mathbf{X}_w\in\mathbb{R}^{is×4}`.\n
    `[reference] <localhost:63342/ivy/docs/source/references/mvg_textbook.pdf#page=174>`_ # noqa
    page 156, equation 6.6

    Parameters
    ----------
    coords_wrt_world
        World-centric homogeneous co-ordinates image *[batch_shape,image_shape,4]*
    ext_mat
        Extrinsic matrix *[batch_shape,3,4]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image shape. Inferred from inputs in None. (Default value = None)
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        Camera-centric homogeneous co-ordinates image *[batch_shape,image_shape,4]*

    """
    if batch_shape is None:
        batch_shape = ext_mat.shape[:-2]
    num_batch_dims = len(batch_shape)

    if image_shape is None:
        image_shape = coords_wrt_world.shape[num_batch_dims:-1]

    if device is None:
        device = ivy.dev(coords_wrt_world)

    # shapes as list
    batch_shape = list(batch_shape)
    image_shape = list(image_shape)

    # BS x IS x 3
    cam_coords = ivy_pg.transform(coords_wrt_world, ext_mat, batch_shape, image_shape)

    # BS x IS x 4
    return ivy.concat(
        (cam_coords, ivy.ones(batch_shape + image_shape + [1], device=device)), axis=-1
    )


def cam_to_world_coords(
    coords_wrt_cam, inv_ext_mat, batch_shape=None, image_shape=None, device=None
):
    r"""Get world-centric homogeneous co-ordinates image
    :math:`\mathbf{X}_w\in\mathbb{R}^{is×4}` from camera-centric
    homogeneous co-ordinates image
    :math:`\mathbf{X}_c\in\mathbb{R}^{is×4}`.\n
    `[reference] <localhost:63342/ivy/docs/source/references/mvg_textbook.pdf#page=174>`_ # noqa
    matrix inverse of page 156, equation 6.6

    Parameters
    ----------
    coords_wrt_cam
        Camera-centric homogeneous co-ordinates image *[batch_shape,image_shape,4]*
    inv_ext_mat
        Inverse extrinsic matrix *[batch_shape,3,4]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image shape. Inferred from inputs in None. (Default value = None)
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        World-centric homogeneous co-ordinates image *[batch_shape,image_shape,4]*

    """
    if batch_shape is None:
        batch_shape = inv_ext_mat.shape[:-2]
    num_batch_dims = len(batch_shape)

    if image_shape is None:
        image_shape = coords_wrt_cam.shape[num_batch_dims:-1]

    if device is None:
        device = ivy.dev(coords_wrt_cam)

    # shapes as list
    batch_shape = list(batch_shape)
    image_shape = list(image_shape)

    # BS x IS x 3
    world_coords = ivy_pg.transform(
        coords_wrt_cam, inv_ext_mat, batch_shape, image_shape
    )

    # BS x IS x 4
    return ivy.concat(
        (world_coords, ivy.ones(batch_shape + image_shape + [1], device=device)),
        axis=-1,
    )


def world_to_ds_pixel_coords(
    coords_wrt_world, full_mat, batch_shape=None, image_shape=None
):
    r"""Get depth scaled homogeneous pixel co-ordinates image
    :math:`\mathbf{X}_p\in\mathbb{R}^{is×3}` from world-centric
    homogeneous co-ordinates image
    :math:`\mathbf{X}_w\in\mathbb{R}^{is×4}`.\n
    `[reference] <localhost:63342/ivy/docs/source/references/mvg_textbook.pdf#page=173>`_ # noqa
    combination of page 156, equation 6.6, and page 155, equation 6.3

    Parameters
    ----------
    coords_wrt_world
        World-centric homogeneous co-ordinates image *[batch_shape,image_shape,4]*
    full_mat
        Full projection matrix *[batch_shape,3,4]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image shape. Inferred from inputs in None. (Default value = None)

    Returns
    -------
    ret
        Depth scaled homogeneous pixel co-ordinates image *[batch_shape,image_shape,3]*

    """
    if batch_shape is None:
        batch_shape = full_mat.shape[:-2]
    num_batch_dims = len(batch_shape)

    if image_shape is None:
        image_shape = coords_wrt_world.shape[num_batch_dims:-1]

    # shapes as list
    batch_shape = list(batch_shape)
    image_shape = list(image_shape)

    # BS x IS x 3
    return ivy_pg.transform(coords_wrt_world, full_mat, batch_shape, image_shape)


def world_coords_to_depth(
    coords_wrt_world, full_mat, batch_shape=None, image_shape=None
):
    r"""Get depth image :math:`\mathbf{X}_d\in\mathbb{R}^{is×1}` from world-centric
    homogeneous co-ordinates image :math:`\mathbf{X}_w\in\mathbb{R}^{is×4}`.\n

    Parameters
    ----------
    coords_wrt_world
        World-centric homogeneous co-ordinates image *[batch_shape,image_shape,4]*
    full_mat
        Full projection matrix *[batch_shape,3,4]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image shape. Inferred from inputs in None. (Default value = None)

    Returns
    -------
    ret
        Depth image *[batch_shape,image_shape,1]*

    """
    # BS x IS x 1
    return world_to_ds_pixel_coords(
        coords_wrt_world, full_mat, batch_shape, image_shape
    )[..., -1:]


def ds_pixel_to_world_coords(
    ds_pixel_coords, inv_full_mat, batch_shape=None, image_shape=None
):
    r"""Get world-centric homogeneous co-ordinates image
    :math:`\mathbf{X}_w\in\mathbb{R}^{is×4}` from depth scaled
    homogeneous pixel co-ordinates image
    :math:`\mathbf{X}_p\in\mathbb{R}^{is×3}`.\n
    `[reference] <localhost:63342/ivy/docs/source/references/mvg_textbook.pdf#page=173>`_ # noqa
    combination of page 155, matrix inverse of equation 6.3,
    and matrix inverse of page 156, equation 6.6

    Parameters
    ----------
    ds_pixel_coords
        Depth scaled homogeneous pixel co-ordinates image: *[batch_shape,image_shape,3]*
    inv_full_mat
        Inverse full projection matrix *[batch_shape,3,4]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image shape. Inferred from inputs in None. (Default value = None)

    Returns
    -------
    ret
        World-centric homogeneous co-ordinates image *[batch_shape,image_shape,4]*

    """
    if batch_shape is None:
        batch_shape = inv_full_mat.shape[:-2]
    num_batch_dims = len(batch_shape)

    if image_shape is None:
        image_shape = ds_pixel_coords.shape[num_batch_dims:-1]

    # shapes as list
    batch_shape = list(batch_shape)
    image_shape = list(image_shape)

    # BS x IS x 4
    ds_pixel_coords = ivy_mech.make_coordinates_homogeneous(
        ds_pixel_coords, batch_shape
    )

    # BS x IS x 3
    world_coords = ivy_pg.transform(
        ds_pixel_coords, inv_full_mat, batch_shape, image_shape
    )

    # BS x IS x 4
    return ivy_mech.make_coordinates_homogeneous(
        world_coords, batch_shape + image_shape
    )


def depth_to_world_coords(
    depth, inv_full_mat, uniform_pixel_coords=None, batch_shape=None, image_dims=None
):
    r"""Get world-centric homogeneous co-ordinates image
    :math:`\mathbf{X}_w\in\mathbb{R}^{hxw×4}` from depth image
    :math:`\mathbf{X}_d\in\mathbb{R}^{hxw×1}`.\n

    Parameters
    ----------
    depth
        Depth image: *[batch_shape,h,w,1]*
    inv_full_mat
        Inverse full projection matrix *[batch_shape,3,4]*
    uniform_pixel_coords
        Image of homogeneous pixel co-ordinates. Created if None.
        *[batch_shape,h,w,3]*
        (Default value = None)
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_dims
        Image shape. Inferred from inputs in None. (Default value = None)

    Returns
    -------
    ret
        World-centric homogeneous co-ordinates image *[batch_shape,h,w,4]*

    """
    if batch_shape is None:
        batch_shape = depth.shape[:-3]

    if image_dims is None:
        image_dims = depth.shape[-3:-1]

    # shapes as list
    batch_shape = list(batch_shape)
    image_dims = list(image_dims)

    # BS x H x W x 3
    ds_pixel_coords = depth_to_ds_pixel_coords(
        depth, uniform_pixel_coords, batch_shape, image_dims
    )

    # BS x H x W x 4
    return ds_pixel_to_world_coords(
        ds_pixel_coords, inv_full_mat, batch_shape, image_dims
    )


def pixel_coords_to_world_ray_vectors(
    inv_full_mat,
    pixel_coords=None,
    camera_center=None,
    batch_shape=None,
    image_shape=None,
):
    r"""Calculate world-centric ray vector image
    :math:`\mathbf{RV}\in\mathbb{R}^{is×3}`
    from homogeneous pixel co-ordinate
    image :math:`\mathbf{X}_p\in\mathbb{R}^{is×3}`. Each ray vector
    :math:`\mathbf{rv}_{i,j}\in\mathbb{R}^{3}` is
    represented as a unit vector from the camera center
    :math:`\overset{\sim}{\mathbf{C}}\in\mathbb{R}^{3×1}`, in the
    world frame. Co-ordinates
    :math:`\mathbf{x}_{i,j}\in\mathbb{R}^{3}` along the world ray
    can then be parameterized as
    :math:`\mathbf{x}_{i,j}=\overset{\sim}{\mathbf{C}} + λ\mathbf{rv}_{i,j}`,
    where
    :math:`λ` is a scalar who's
    magnitude dictates the position of the world co-ordinate along the world ray.

    Parameters
    ----------
    inv_full_mat
        Inverse full projection matrix *[batch_shape,3,4]*
    pixel_coords
        Homogeneous pixel co-ordinates image, created uniformly if None.
        *[batch_shape,image_shape,3]* (Default value = None)
    camera_center
        Camera centers, inferred from inv_full_mat if None *[batch_shape,3,1]*
        (Default value = None)
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image shape. Inferred from inputs in None. (Default value = None)

    Returns
    -------
    ret
        World ray vectors *[batch_shape,image_shape,3]*

    """
    if batch_shape is None:
        batch_shape = inv_full_mat.shape[:-2]
    num_batch_dims = len(batch_shape)

    if image_shape is None:
        if pixel_coords is None:
            raise Exception(
                "if pixel_coords is not specified, "
                "image_shape must be specified when calling"
                "pixel_coords_to_world_ray_vectors"
            )
        image_shape = pixel_coords.shape[num_batch_dims:-1]
    num_image_dims = len(image_shape)

    # shapes as list
    batch_shape = list(batch_shape)
    image_shape = list(image_shape)

    if camera_center is None:
        camera_center = inv_ext_mat_to_camera_center(inv_full_mat)

    if pixel_coords is None:
        pixel_coords = create_uniform_pixel_coords_image(
            image_shape, batch_shape, device=ivy.dev(inv_full_mat)
        )

    # BS x [1]xNID x 3
    camera_centers_reshaped = ivy.reshape(
        camera_center, batch_shape + [1] * num_image_dims + [3]
    )

    # BS x IS x 3
    vectors = (
        ds_pixel_to_world_coords(pixel_coords, inv_full_mat, batch_shape, image_shape)[
            ..., 0:3
        ]
        - camera_centers_reshaped
    )

    # BS x H x W x 3
    return vectors / (
        ivy.sum(vectors**2, axis=-1, keepdims=True) ** 0.5 + MIN_DENOMINATOR
    )


def sphere_coords_to_world_ray_vectors(
    sphere_coords, inv_rotation_mat, batch_shape=None, image_shape=None
):
    r"""Calculate world-centric ray vector image
    :math:`\mathbf{RV}\in\mathbb{R}^{is×3}` from camera-centric ego-sphere
    polar co-ordinates image :math:`\mathbf{S}_c\in\mathbb{R}^{is×3}`.
    Each ray vector
    :math:`\mathbf{rv}_{i,j}\in\mathbb{R}^{3}` is represented as
    a unit vector from the camera center
    :math:`\overset{\sim}{\mathbf{C}}\in\mathbb{R}^{3×1}`,
    in the world frame.
    Co-ordinates
    :math:`\mathbf{x}_{i,j}\in\mathbb{R}^{3}` along the world ray can
    then be parameterized as
    :math:`\mathbf{x}_{i,j}=\overset{\sim}{\mathbf{C}} + λ\mathbf{rv}_{i,j}`,
    where
    :math:`λ` is a scalar who's
    magnitude dictates the position of the world co-ordinate along the world ray.

    Parameters
    ----------
    sphere_coords
        Camera-centric ego-sphere polar co-ordinates image *[batch_shape,image_shape,3]*
    inv_rotation_mat
        Inverse rotation matrix *[batch_shape,3,3]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image dimensions. Inferred from inputs in None. (Default value = None)

    Returns
    -------
    ret
        World ray vectors *[batch_shape,image_shape,3]*

    """
    if batch_shape is None:
        batch_shape = inv_rotation_mat.shape[:-2]
    num_batch_dims = len(batch_shape)

    if image_shape is None:
        image_shape = sphere_coords.shape[num_batch_dims:-1]

    # shapes as list
    batch_shape = list(batch_shape)
    image_shape = list(image_shape)

    # BS x IS x 3
    cam_coords = sphere_to_cam_coords(
        sphere_coords, batch_shape=batch_shape + image_shape
    )[..., 0:3]
    vectors = ivy_pg.transform(cam_coords, inv_rotation_mat, batch_shape, image_shape)
    return vectors / (
        ivy.sum(vectors**2, axis=-1, keepdims=True) ** 0.5 + MIN_DENOMINATOR
    )


def bilinearly_interpolate_image(
    image, sampling_pixel_coords, batch_shape=None, image_dims=None
):
    r"""Bilinearly interpolate image
    :math:`\mathbf{X}\in\mathbb{R}^{h×w×d}` at sampling pixel locations
    :math:`\mathbf{S}\in\mathbb{R}^{h×w×2}`, to return interpolated image
    :math:`\mathbf{X}_I\in\mathbb{R}^{h×w×d}`.\n
    `[reference] <https://en.wikipedia.org/wiki/Bilinear_interpolation>`_

    Parameters
    ----------
    image
        Image to be interpolated *[batch_shape,h,w,d]*
    sampling_pixel_coords
        Pixel co-ordinates to sample the image at *[batch_shape,h,w,2]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_dims
        Image dimensions. Inferred from inputs in None. (Default value = None)

    Returns
    -------
    ret
        Interpolated image *[batch_shape,h,w,d]*

    """
    if batch_shape is None:
        batch_shape = image.shape[:-3]

    if image_dims is None:
        image_dims = image.shape[-3:-1]

    # shapes as list
    batch_shape = list(batch_shape)
    image_dims = list(image_dims)

    # batch shape product
    batch_shape_product = reduce(mul, batch_shape, 1)

    # prod(BS) x H x W x D
    uniform_values_flat = ivy.reshape(image, [batch_shape_product] + image_dims + [-1])

    # prod(BS) x H x W x 2
    sampling_pixel_coords_flat = ivy.reshape(
        sampling_pixel_coords, [batch_shape_product] + image_dims + [2]
    )

    # prod(BS) x H x W x D
    interpolation_flat = ivy_img.bilinear_resample(
        uniform_values_flat, sampling_pixel_coords_flat
    )

    # BS x H x W x D
    return ivy.reshape(interpolation_flat, batch_shape + image_dims + [-1])


# noinspection PyUnusedLocal
def inv_ext_mat_to_camera_center(inv_ext_mat):
    r"""Compute camera center
    :math:`\overset{\sim}{\mathbf{C}}\in\mathbb{R}^{3×1}`
    from camera extrinsic matrix
    :math:`\mathbf{E}\in\mathbb{R}^{3×4}`.\n
    `[reference] <localhost:63342/ivy/docs/source/references/mvg_textbook.pdf#page=174>`_ # noqa
    matrix inverse of page 156, equation 6.6

    Parameters
    ----------
    inv_ext_mat
        Inverse extrinsic matrix *[batch_shape,3,4]*

    Returns
    -------
    ret
        Camera center *[batch_shape,3,1]*

    """
    # BS x 3 x 1
    return inv_ext_mat[..., -1:]


def calib_and_ext_to_full_mat(calib_mat, ext_mat):
    r"""Compute full projection matrix :math:`\mathbf{P}\in\mathbb{R}^{3×4}`
    from calibration
    :math:`\mathbf{K}\in\mathbb{R}^{3×3}` and extrinsic matrix
    :math:`\mathbf{E}\in\mathbb{R}^{3×4}`.\n

    Parameters
    ----------
    calib_mat
        Calibration matrix *[batch_shape,3,3]*
    ext_mat
        Extrinsic matrix *[batch_shape,3,4]*

    Returns
    -------
    ret
        Full projection matrix *[batch_shape,3,4]*

    """
    # BS x 3 x 4
    return ivy.matmul(calib_mat, ext_mat)


def cam_to_sphere_coords(cam_coords, forward_facing_z=True):
    r"""Convert camera-centric homogeneous cartesian co-ordinates image
    :math:`\mathbf{X}_c\in\mathbb{R}^{h×w×4}` to
    camera-centric ego-sphere polar co-ordinates image
    :math:`\mathbf{S}_c\in\mathbb{R}^{h×w×3}`.\n
    `[reference] <https://en.wikipedia.org/wiki/Spherical_coordinate_system#Cartesian_coordinates>`_ # noqa

    Parameters
    ----------
    cam_coords
        Camera-centric homogeneous cartesian co-ordinates image *[batch_shape,h,w,4]*
    forward_facing_z
        Whether to use reference frame so z is forward facing. Default is False.

    Returns
    -------
    ret
        Camera-centric ego-sphere polar co-ordinates image *[batch_shape,h,w,3]*

    """
    # BS x H x W x 3
    if forward_facing_z:
        cam_coords = ivy.concat(
            (cam_coords[..., 2:3], cam_coords[..., 0:1], cam_coords[..., 1:2]), axis=-1
        )
    else:
        cam_coords = cam_coords[..., 0:3]

    # BS x H x W x 3
    return ivy_mech.cartesian_to_polar_coords(cam_coords)


def ds_pixel_to_sphere_coords(
    ds_pixel_coords, inv_calib_mat, batch_shape=None, image_shape=None
):
    r"""Convert depth scaled homogeneous pixel co-ordinates image
    :math:`\mathbf{X}_p\in\mathbb{R}^{is×3}` to
    camera-centric ego-sphere polar co-ordinates image
    :math:`\mathbf{S}_c\in\mathbb{R}^{is×3}`.\n
    `[reference] <https://en.wikipedia.org/wiki/Spherical_coordinate_system#Cartesian_coordinates>`_ # noqa

    Parameters
    ----------
    ds_pixel_coords
        Depth scaled homogeneous pixel co-ordinates image *[batch_shape,image_shape,3]*
    inv_calib_mat
        Inverse calibration matrix *[batch_shape,3,3]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image dimensions. Inferred from inputs in None. (Default value = None)

    Returns
    -------
    ret
        Camera-centric ego-sphere polar co-ordinates image *[batch_shape,image_shape,3]*

    """
    if batch_shape is None:
        batch_shape = inv_calib_mat.shape[:-2]
    num_batch_dims = len(batch_shape)

    if image_shape is None:
        image_shape = ds_pixel_coords.shape[num_batch_dims:-1]

    # shapes as list
    batch_shape = list(batch_shape)
    image_shape = list(image_shape)

    # BS x IS x 4
    cam_coords = ds_pixel_to_cam_coords(
        ds_pixel_coords, inv_calib_mat, batch_shape, image_shape
    )

    # BS x IS x 3
    return cam_to_sphere_coords(cam_coords)


def angular_pixel_to_sphere_coords(angular_pixel_coords, pixels_per_degree):
    r"""Convert angular pixel co-ordinates image
    :math:`\mathbf{A}_p\in\mathbb{R}^{h×w×3}` to camera-centric ego-sphere
    polar co-ordinates image :math:`\mathbf{S}_c\in\mathbb{R}^{h×w×3}`.\n
    `[reference] <https://en.wikipedia.org/wiki/Spherical_coordinate_system#Cartesian_coordinates>`_ # noqa

    Parameters
    ----------
    angular_pixel_coords
        Angular pixel co-ordinates image *[batch_shape,h,w,3]*
    pixels_per_degree
        Number of pixels per angular degree

    Returns
    -------
    ret
        Camera-centric ego-sphere polar co-ordinates image *[batch_shape,h,w,3]*

    """
    # BS x H x W x 1
    sphere_x_coords = angular_pixel_coords[..., 0:1]
    sphere_y_coords = angular_pixel_coords[..., 1:2]
    radius_values = angular_pixel_coords[..., 2:3]

    sphere_x_angle_coords_in_degs = (
        180 - sphere_x_coords / (pixels_per_degree + MIN_DENOMINATOR) % 360
    )
    sphere_y_angle_coords_in_degs = (
        sphere_y_coords / (pixels_per_degree + MIN_DENOMINATOR) % 180
    )

    # BS x H x W x 2
    sphere_angle_coords_in_degs = ivy.concat(
        (sphere_x_angle_coords_in_degs, sphere_y_angle_coords_in_degs), axis=-1
    )
    sphere_angle_coords = sphere_angle_coords_in_degs * np.pi / 180

    # BS x H x W x 3
    return ivy.concat((sphere_angle_coords, radius_values), axis=-1)


def sphere_to_cam_coords(
    sphere_coords, forward_facing_z=True, batch_shape=None, device=None
):
    r"""Convert camera-centric ego-sphere polar co-ordinates image
    :math:`\mathbf{S}_c\in\mathbb{R}^{bs×3}` to
    camera-centric homogeneous cartesian co-ordinates image
    :math:`\mathbf{X}_c\in\mathbb{R}^{bs×4}`.\n
    `[reference] <https://en.wikipedia.org/wiki/Spherical_coordinate_system#Cartesian_coordinates>`_ # noqa

    Parameters
    ----------
    sphere_coords
        Camera-centric ego-sphere polar co-ordinates image *[batch_shape,3]*
    forward_facing_z
        Whether to use reference frame so z is forward facing. Default is False.
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        Camera-centric homogeneous cartesian co-ordinates image *[batch_shape,4]*

    """
    if batch_shape is None:
        batch_shape = sphere_coords.shape[:-1]

    if device is None:
        device = ivy.dev(sphere_coords)

    # shapes as list
    batch_shape = list(batch_shape)

    # BS x 3
    cam_coords = ivy_mech.polar_to_cartesian_coords(sphere_coords)
    if forward_facing_z:
        cam_coords = ivy.concat(
            (cam_coords[..., 1:2], cam_coords[..., 2:3], cam_coords[..., 0:1]), axis=-1
        )

    # BS x 4
    return ivy_mech.make_coordinates_homogeneous(cam_coords, batch_shape)


def sphere_to_ds_pixel_coords(
    sphere_coords, calib_mat, batch_shape=None, image_shape=None
):
    r"""Convert camera-centric ego-sphere polar co-ordinates image
    :math:`\mathbf{S}_c\in\mathbb{R}^{is×3}` to depth scaled
    homogeneous pixel co-ordinates image :math:`\mathbf{X}_p\in\mathbb{R}^{is×3}`.\n
    `[reference] <https://en.wikipedia.org/wiki/Spherical_coordinate_system#Cartesian_coordinates>`_ # noqa

    Parameters
    ----------
    sphere_coords
        Camera-centric ego-sphere polar co-ordinates image *[batch_shape,image_shape,3]*
    calib_mat
        Calibration matrix *[batch_shape,3,3]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image dimensions. Inferred from inputs in None. (Default value = None)

    Returns
    -------
    ret
        Depth scaled homogeneous pixel co-ordinates image *[batch_shape,image_shape,3]*

    """
    if batch_shape is None:
        batch_shape = calib_mat.shape[:-2]
    num_batch_dims = len(batch_shape)

    if image_shape is None:
        image_shape = sphere_coords.shape[num_batch_dims:-1]

    # shapes as list
    batch_shape = list(batch_shape)
    image_shape = list(image_shape)

    # BS x IS x 4
    cam_coords = sphere_to_cam_coords(
        sphere_coords, batch_shape=batch_shape + image_shape
    )

    # BS x IS x 3
    return cam_to_ds_pixel_coords(cam_coords, calib_mat, batch_shape, image_shape)


def sphere_to_angular_pixel_coords(sphere_coords, pixels_per_degree):
    r"""Convert camera-centric ego-sphere polar co-ordinates image
    :math:`\mathbf{S}_c\in\mathbb{R}^{h×w×3}` to angular
    pixel co-ordinates image :math:`\mathbf{A}_p\in\mathbb{R}^{h×w×3}`.\n
    `[reference] <https://en.wikipedia.org/wiki/Spherical_coordinate_system#Cartesian_coordinates>`_ # noqa

    Parameters
    ----------
    sphere_coords
        Camera-centric ego-sphere polar co-ordinates image *[batch_shape,h,w,3]*
    pixels_per_degree
        Number of pixels per angular degree

    Returns
    -------
    ret
        Angular pixel co-ordinates image *[batch_shape,h,w,3]*

    """
    # BS x H x W x 1
    sphere_radius_vals = sphere_coords[..., -1:]

    # BS x H x W x 2
    sphere_angle_coords = sphere_coords[..., 0:2]

    # BS x H x W x 2
    sphere_angle_coords_in_degs = sphere_angle_coords * 180 / np.pi

    # BS x H x W x 1
    sphere_x_coords = (
        (180 - sphere_angle_coords_in_degs[..., 0:1]) % 360
    ) * pixels_per_degree
    sphere_y_coords = (sphere_angle_coords_in_degs[..., 1:2] % 180) * pixels_per_degree

    # BS x H x W x 3
    return ivy.concat((sphere_x_coords, sphere_y_coords, sphere_radius_vals), axis=-1)


# Camera Geometry Object Functions #
# ---------------------------------#


def persp_angles_and_pp_offsets_to_intrinsics_object(
    persp_angles, pp_offsets, image_dims, batch_shape=None
):
    """Create camera intrinsics object from perspective angles :math:`θ_x, θ_y`,
    principal-point offsets :math:`p_x, p_y`
    and image dimensions [height, width].

    Parameters
    ----------
    persp_angles
        Perspective angles *[batch_shape,2]*
    pp_offsets
        Principal-point offsets *[batch_shape,2]*
    image_dims
        Image dimensions.
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)

    Returns
    -------
    ret
        Camera intrinsics object.

    """
    if batch_shape is None:
        batch_shape = persp_angles.shape[:-1]

    # shapes as list
    batch_shape = list(batch_shape)
    image_dims = list(image_dims)

    # BS x 2
    focal_lengths = persp_angles_to_focal_lengths(persp_angles, image_dims)

    # BS x 3 x 3
    calib_mat = focal_lengths_and_pp_offsets_to_calib_mat(
        focal_lengths, pp_offsets, batch_shape
    )

    # BS x 3 x 3
    inv_calib_mat = ivy.inv(calib_mat)

    # intrinsics object
    intrinsics = Intrinsics(
        focal_lengths, persp_angles, pp_offsets, calib_mat, inv_calib_mat
    )
    return intrinsics


def focal_lengths_and_pp_offsets_to_intrinsics_object(
    focal_lengths, pp_offsets, image_dims, batch_shape=None
):
    """Create camera intrinsics object from focal lengths :math:`f_x, f_y`,
    principal-point offsets :math:`p_x, p_y`, and
    image dimensions [height, width].

    Parameters
    ----------
    focal_lengths
        Focal lengths *[batch_shape,2]*
    pp_offsets
        Principal-point offsets *[batch_shape,2]*
    image_dims
        Image dimensions. Inferred from inputs in None.
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)

    Returns
    -------
    ret
        Camera intrinsics object

    """
    if batch_shape is None:
        batch_shape = focal_lengths.shape[:-1]

    # shapes as list
    batch_shape = list(batch_shape)
    image_dims = list(image_dims)

    # BS x 2
    persp_angles = focal_lengths_to_persp_angles(focal_lengths, image_dims)

    # BS x 3 x 3
    calib_mat = focal_lengths_and_pp_offsets_to_calib_mat(
        focal_lengths, pp_offsets, batch_shape
    )

    # BS x 3 x 3
    inv_calib_mat = ivy.inv(calib_mat)

    # intrinsics object
    intrinsics = Intrinsics(
        focal_lengths, persp_angles, pp_offsets, calib_mat, inv_calib_mat
    )
    return intrinsics


def calib_mat_to_intrinsics_object(calib_mat, image_dims, batch_shape=None):
    """Create camera intrinsics object from calibration matrix.

    Parameters
    ----------
    calib_mat
        Calibration matrices *[batch_shape,3,3]*
    image_dims
        Image dimensions. Inferred from inputs in None.
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)

    Returns
    -------
    ret
        Camera intrinsics object

    """
    if batch_shape is None:
        batch_shape = calib_mat.shape[:-2]

    # shapes as list
    batch_shape = list(batch_shape)
    image_dims = list(image_dims)

    # BS x 2
    focal_lengths = ivy.concat(
        (calib_mat[..., 0, 0:1], calib_mat[..., 1, 1:2]), axis=-1
    )

    # BS x 2
    persp_angles = focal_lengths_to_persp_angles(focal_lengths, image_dims)

    # BS x 2
    pp_offsets = ivy.concat((calib_mat[..., 0, -1:], calib_mat[..., 1, -1:]), axis=-1)

    # BS x 3 x 3
    inv_calib_mat = ivy.inv(calib_mat)

    # intrinsics object
    intrinsics = Intrinsics(
        focal_lengths, persp_angles, pp_offsets, calib_mat, inv_calib_mat
    )
    return intrinsics


def ext_mat_and_intrinsics_to_cam_geometry_object(
    ext_mat, intrinsics, batch_shape=None, device=None
):
    r"""Create camera geometry object from extrinsic matrix
    :math:`\mathbf{E}\in\mathbb{R}^{3×4}`, and camera intrinsics
    object.

    Parameters
    ----------
    ext_mat
        Extrinsic matrix *[batch_shape,3,4]*
    intrinsics
        camera intrinsics object
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        Camera geometry object

    """
    if batch_shape is None:
        batch_shape = ext_mat.shape[:-2]

    if device is None:
        device = ivy.dev(ext_mat)

    # shapes as list
    batch_shape = list(batch_shape)

    # num batch dims
    num_batch_dims = len(batch_shape)

    # BS x 4 x 4
    ext_mat_homo = ivy.concat(
        (
            ext_mat,
            ivy.tile(
                ivy.reshape(
                    ivy.array([0.0, 0.0, 0.0, 1.0], device=device),
                    [1] * (num_batch_dims + 1) + [4],
                ),
                batch_shape + [1, 1],
            ),
        ),
        axis=-2,
    )

    # BS x 4 x 4
    inv_ext_mat_homo = ivy.inv(ext_mat_homo)

    # BS x 3 x 4
    inv_ext_mat = inv_ext_mat_homo[..., 0:3, :]

    # BS x 3 x 1
    cam_center = inv_ext_mat_to_camera_center(inv_ext_mat)

    # BS x 3 x 3
    Rs = ext_mat[..., 0:3]

    # BS x 3 x 3
    inv_Rs = inv_ext_mat[..., 0:3]

    # extrinsics object
    extrinsics = Extrinsics(cam_center, Rs, inv_Rs, ext_mat_homo, inv_ext_mat_homo)

    # BS x 3 x 4
    full_mat = calib_and_ext_to_full_mat(intrinsics.calib_mats, ext_mat)

    # BS x 4 x 4
    full_mat_homo = ivy.concat(
        (
            full_mat,
            ivy.tile(
                ivy.reshape(
                    ivy.array([0.0, 0.0, 0.0, 1.0], device=device),
                    [1] * (num_batch_dims + 1) + [4],
                ),
                batch_shape + [1, 1],
            ),
        ),
        axis=-2,
    )

    # BS x 4 x 4
    inv_full_mat_homo = ivy.inv(full_mat_homo)

    # camera geometry object
    cam_geometry = CameraGeometry(
        intrinsics, extrinsics, full_mat_homo, inv_full_mat_homo
    )
    return cam_geometry


def inv_ext_mat_and_intrinsics_to_cam_geometry_object(
    inv_ext_mat, intrinsics, batch_shape=None, device=None
):
    r"""Create camera geometry object from inverse extrinsic matrix
    :math:`\mathbf{E}^{-1}\in\mathbb{R}^{3×4}`, and camera
    intrinsics object.

    Parameters
    ----------
    inv_ext_mat
        Inverse extrinsic matrix *[batch_shape,3,4]*
    intrinsics
        camera intrinsics object
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        Camera geometry object

    """
    if batch_shape is None:
        batch_shape = inv_ext_mat.shape[:-2]

    if device is None:
        device = ivy.dev(inv_ext_mat)

    # shapes as list
    batch_shape = list(batch_shape)

    # num batch dims
    num_batch_dims = len(batch_shape)

    # BS x 4 x 4
    inv_ext_mat_homo = ivy.concat(
        (
            inv_ext_mat,
            ivy.tile(
                ivy.reshape(
                    ivy.array([0.0, 0.0, 0.0, 1.0], device=device),
                    [1] * (num_batch_dims + 1) + [4],
                ),
                batch_shape + [1, 1],
            ),
        ),
        axis=-2,
    )

    # BS x 4 x 4
    ext_mat_homo = ivy.inv(inv_ext_mat_homo)

    # BS x 3 x 4
    ext_mat = ext_mat_homo[..., 0:3, :]

    # BS x 3 x 1
    cam_center = inv_ext_mat_to_camera_center(inv_ext_mat)

    # BS x 3 x 3
    Rs = ext_mat[..., 0:3]

    # BS x 3 x 3
    inv_Rs = inv_ext_mat[..., 0:3]

    # extrinsics object
    extrinsics = Extrinsics(cam_center, Rs, inv_Rs, ext_mat_homo, inv_ext_mat_homo)

    # BS x 3 x 4
    full_mat = calib_and_ext_to_full_mat(intrinsics.calib_mats, ext_mat)

    # BS x 4 x 4
    full_mat_homo = ivy.concat(
        (
            full_mat,
            ivy.tile(
                ivy.reshape(
                    ivy.array([0.0, 0.0, 0.0, 1.0], device=device),
                    [1] * (num_batch_dims + 1) + [4],
                ),
                batch_shape + [1, 1],
            ),
        ),
        axis=-2,
    )

    # BS x 4 x 4
    inv_full_mat_homo = ivy.inv(full_mat_homo)

    # camera geometry object
    camera_geometry = CameraGeometry(
        intrinsics, extrinsics, full_mat_homo, inv_full_mat_homo
    )
    return camera_geometry



"""Collection of Padding Functions"""

# global
import ivy


def pad_omni_image(image, pad_size, image_dims=None):
    """Pad an omni-directional image with the correct image wrapping at the edges.

    Parameters
    ----------
    image
        Image to perform the padding on *[batch_shape,h,w,d]*
    pad_size
        Number of pixels to pad.
    image_dims
        Image dimensions. Inferred from Inputs if None. (Default value = None)

    Returns
    -------
    ret
        New padded omni-directional image *[batch_shape,h+ps,w+ps,d]*

    """
    if image_dims is None:
        image_dims = image.shape[-3:-1]

    # BS x PS x W/2 x D
    top_left = image[..., 0:pad_size, int(image_dims[1] / 2) :, :]
    top_right = image[..., 0:pad_size, 0 : int(image_dims[1] / 2), :]

    # BS x PS x W x D
    top_border = ivy.flip(ivy.concat((top_left, top_right), axis=-2), axis=-3)

    # BS x PS x W/2 x D
    bottom_left = image[..., -pad_size:, int(image_dims[1] / 2) :, :]
    bottom_right = image[..., -pad_size:, 0 : int(image_dims[1] / 2), :]

    # BS x PS x W x D
    bottom_border = ivy.flip(ivy.concat((bottom_left, bottom_right), axis=-2), axis=-3)

    # BS x H+2PS x W x D
    image_expanded = ivy.concat((top_border, image, bottom_border), axis=-3)

    # BS x H+2PS x PS x D
    left_border = image_expanded[..., -pad_size:, :]
    right_border = image_expanded[..., 0:pad_size, :]

    # BS x H+2PS x W+2PS x D
    return ivy.concat((left_border, image_expanded, right_border), axis=-2)



"""Collection of Two-View-Geometry Functions"""

# global
import ivy
import ivy_mech
from operator import mul
from functools import reduce

# local
from ivy_vision import projective_geometry as ivy_pg
from ivy_vision import single_view_geometry as ivy_svg

MIN_DENOMINATOR = 1e-12


def ds_pixel_to_ds_pixel_coords(
    ds_pixel_coords1, cam1to2_full_mat, batch_shape=None, image_shape=None, device=None
):
    r"""Transform depth scaled homogeneous pixel co-ordinates image in
    first camera frame
    :math:`\mathbf{X}_{p1}\in\mathbb{R}^{is×3}` to
    depth scaled homogeneous pixel co-ordinates image in second camera
    frame :math:`\mathbf{X}_{p2}\in\mathbb{R}^{is×3}`,
    given camera to camera projection matrix
    :math:`\mathbf{P}_{1→2}\in\mathbb{R}^{3×4}`.\n
    `[reference] <localhost:63342/ivy/docs/source/references/mvg_textbook.pdf#page=174>`_ # noqa

    Parameters
    ----------
    ds_pixel_coords1
        Depth scaled homogeneous pixel co-ordinates image in
        frame 1 *[batch_shape,image_shape,3]*
    cam1to2_full_mat
        Camera1-to-camera2 full projection matrix *[batch_shape,3,4]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image shape. Inferred from inputs in None. (Default value = None)
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        Depth scaled homogeneous pixel co-ordinates image in frame 2
        *[batch_shape,image_shape,3]*

    """
    if batch_shape is None:
        batch_shape = cam1to2_full_mat.shape[:-2]
    num_batch_dims = len(batch_shape)

    if image_shape is None:
        image_shape = ds_pixel_coords1.shape[num_batch_dims:-1]

    if device is None:
        device = ivy.dev(ds_pixel_coords1)

    # shapes as list
    batch_shape = list(batch_shape)
    image_shape = list(image_shape)

    # BS x IS x 4
    pixel_coords_homo = ivy_mech.make_coordinates_homogeneous(ds_pixel_coords1)

    # BS x IS x 3
    return ivy_pg.transform(
        pixel_coords_homo, cam1to2_full_mat, batch_shape, image_shape
    )


def cam_to_cam_coords(
    cam_coords1, cam1to2_ext_mat, batch_shape=None, image_shape=None, device=None
):
    r"""Transform camera-centric homogeneous co-ordinates image for camera 1
    :math:`\mathbf{X}_{c1}\in\mathbb{R}^{is×4}` to
    camera-centric homogeneous co-ordinates image for camera 2 :
    math:`\mathbf{X}_{c2}\in\mathbb{R}^{is×4}`.\n
    `[reference] <localhost:63342/ivy/docs/source/references/mvg_textbook.pdf#page=174>`_ # noqa

    Parameters
    ----------
    cam_coords1
        Camera-centric homogeneous co-ordinates image in
        frame 1 *[batch_shape,image_shape,4]*
    cam1to2_ext_mat
        Camera1-to-camera2 extrinsic projection matrix *[batch_shape,3,4]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image shape. Inferred from inputs in None. (Default value = None)
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        Depth scaled homogeneous pixel co-ordinates image
        in frame 2 *[batch_shape,image_shape,3]*

    """
    if batch_shape is None:
        batch_shape = cam1to2_ext_mat.shape[:-2]
    num_batch_dims = len(batch_shape)

    if image_shape is None:
        image_shape = cam_coords1.shape[num_batch_dims:-1]

    if device is None:
        device = ivy.dev(cam_coords1)

    # shapes as list
    batch_shape = list(batch_shape)
    image_shape = list(image_shape)

    # BS x IS x 3
    cam_coords2 = ivy_pg.transform(
        cam_coords1, cam1to2_ext_mat, batch_shape, image_shape
    )

    # BS x IS x 4
    return ivy_mech.make_coordinates_homogeneous(cam_coords2)


def sphere_to_sphere_coords(
    sphere_coords1, cam1to2_ext_mat, batch_shape=None, image_shape=None
):
    r"""Convert camera-centric ego-sphere polar co-ordinates image in frame 1
    :math:`\mathbf{S}_{c1}\in\mathbb{R}^{is×3}`
    to camera-centric ego-sphere polar co-ordinates image in frame 2
    :math:`\mathbf{S}_{c2}\in\mathbb{R}^{is×3}`.\n
    `[reference] <https://en.wikipedia.org/wiki/Spherical_coordinate_system#Cartesian_coordinates>`_ # noqa

    Parameters
    ----------
    sphere_coords1
        Camera-centric ego-sphere polar co-ordinates image in frame 1
        *[batch_shape,image_shape,3]*
    cam1to2_ext_mat
        Camera1-to-camera2 extrinsic projection matrix *[batch_shape,3,4]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image shape. Inferred from inputs in None. (Default value = None)

    Returns
    -------
    ret
        Camera-centric ego-sphere polar co-ordinates image in frame 2
        *[batch_shape,image_shape,3]*

    """
    if batch_shape is None:
        batch_shape = cam1to2_ext_mat.shape[:-2]
    num_batch_dims = len(batch_shape)

    if image_shape is None:
        image_shape = sphere_coords1.shape[num_batch_dims:-1]

    # shapes as list
    batch_shape = list(batch_shape)
    image_shape = list(image_shape)

    # BS x IS x 4
    cam_coords1 = ivy_svg.sphere_to_cam_coords(
        sphere_coords1, batch_shape=batch_shape + image_shape
    )
    cam_coords2 = cam_to_cam_coords(
        cam_coords1, cam1to2_ext_mat, batch_shape, image_shape
    )

    # BS x IS x 3
    return ivy_svg.cam_to_sphere_coords(cam_coords2)


def angular_pixel_to_angular_pixel_coords(
    angular_pixel_coords1,
    cam1to2_ext_mat,
    pixels_per_degree,
    batch_shape=None,
    image_shape=None,
):
    r"""Convert angular pixel co-ordinates image in frame 1
    :math:`\mathbf{A}_{p1}\in\mathbb{R}^{is×3}` to angular pixel
    co-ordinates image in frame 2 :math:`\mathbf{A}_{p2}\in\mathbb{R}^{is×3}`.\n
    `[reference] <https://en.wikipedia.org/wiki/Spherical_coordinate_system#Cartesian_coordinates>`_ # noqa

    Parameters
    ----------
    angular_pixel_coords1
        Angular pixel co-ordinates image in frame 1 *[batch_shape,image_shape,3]*
    cam1to2_ext_mat
        Camera1-to-camera2 extrinsic projection matrix *[batch_shape,3,4]*
    pixels_per_degree
        Number of pixels per angular degree
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image dimensions. Inferred from inputs in None. (Default value = None)

    Returns
    -------
    ret
        Camera-centric ego-sphere polar co-ordinates image in frame 2
        *[batch_shape,image_shape,3]*

    """
    if batch_shape is None:
        batch_shape = cam1to2_ext_mat.shape[:-2]
    num_batch_dims = len(batch_shape)

    if image_shape is None:
        image_shape = angular_pixel_coords1.shape[num_batch_dims:-1]

    # shapes as list
    batch_shape = list(batch_shape)
    image_shape = list(image_shape)

    # BS x IS x 3
    sphere_coords1 = ivy_svg.angular_pixel_to_sphere_coords(
        angular_pixel_coords1, pixels_per_degree
    )

    # BS x IS x 3
    sphere_coords2 = sphere_to_sphere_coords(
        sphere_coords1, cam1to2_ext_mat, batch_shape, image_shape
    )

    # BS x IS x 3
    return ivy_svg.sphere_to_angular_pixel_coords(sphere_coords2, pixels_per_degree)


def get_fundamental_matrix(
    full_mat1,
    full_mat2,
    camera_center1=None,
    pinv_full_mat1=None,
    batch_shape=None,
    device=None,
):
    r"""Compute fundamental matrix
    :math:`\mathbf{F}\in\mathbb{R}^{3×3}` between two cameras,
    given their extrinsic
    matrices :math:`\mathbf{E}_1\in\mathbb{R}^{3×4}` and
    :math:`\mathbf{E}_2\in\mathbb{R}^{3×4}`.\n
    `[reference] <localhost:63342/ivy/docs/source/references/mvg_textbook.pdf#page=262>`_ # noqa
    bottom of page 244, section 9.2.2, equation 9.1

    Parameters
    ----------
    full_mat1
        Frame 1 full projection matrix *[batch_shape,3,4]*
    full_mat2
        Frame 2 full projection matrix *[batch_shape,3,4]*
    camera_center1
        Frame 1 camera center, inferred from full_mat1 if None *[batch_shape,3,1]*
        (Default value = None)
    pinv_full_mat1
        Frame 1 full projection matrix pseudo-inverse, inferred from full_mat1 if None
        *[batch_shape,4,3]* (Default value = None)
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        Fundamental matrix connecting frames 1 and 2 *[batch_shape,3,3]*

    """
    if batch_shape is None:
        batch_shape = full_mat1.shape[:-2]

    if device is None:
        device = ivy.dev(full_mat1)

    # shapes as list
    batch_shape = list(batch_shape)

    if camera_center1 is None:
        inv_full_mat1 = ivy.inv(
            ivy_mech.make_transformation_homogeneous(full_mat1, batch_shape, device)
        )[..., 0:3, :]
        camera_center1 = ivy_svg.inv_ext_mat_to_camera_center(inv_full_mat1)

    if pinv_full_mat1 is None:
        pinv_full_mat1 = ivy.pinv(full_mat1)

    # BS x 4 x 1
    camera_center1_homo = ivy.concat(
        (camera_center1, ivy.ones(batch_shape + [1, 1], device=device)), axis=-2
    )

    # BS x 3
    e2 = ivy.matmul(full_mat2, camera_center1_homo)[..., -1]

    # BS x 3 x 3
    e2_skew_symmetric = ivy.vector_to_skew_symmetric_matrix(e2)

    # BS x 3 x 3
    return ivy.matmul(e2_skew_symmetric, ivy.matmul(full_mat2, pinv_full_mat1))


# noinspection PyUnresolvedReferences
def closest_mutual_points_along_two_skew_rays(
    camera_centers, world_ray_vectors, batch_shape=None, image_shape=None, device=None
):
    r"""Compute closest mutual homogeneous co-ordinates
    :math:`\mathbf{x}_{1,i,j}\in\mathbb{R}^{4}` and
    :math:`\mathbf{x}_{2,i,j}\in\mathbb{R}^{4}` along two world-centric rays
    :math:`\overset{\sim}{\mathbf{C}_1} + λ_1\mathbf{rv}_{1,i,j}` and
    :math:`\overset{\sim}{\mathbf{C}_2} + λ_2\mathbf{rv}_{2,i,j}`,
    for each index aligned pixel between two
    world-centric ray vector images
    :math:`\mathbf{RV}_1\in\mathbb{R}^{is×3}` and
    :math:`\mathbf{RV}_2\in\mathbb{R}^{is×3}`.
    The function returns two images of closest mutual homogeneous
    co-ordinates :math:`\mathbf{X}_1\in\mathbb{R}^{is×4}` and
    :math:`\mathbf{X}_2\in\mathbb{R}^{is×4}`,
    concatenated together into a single array.\n
    `[reference] <https://math.stackexchange.com/questions/1414285/location-of-shortest-distance-between-two-skew-lines-in-3d>`_ # noqa
    second answer in forum

    Parameters
    ----------
    camera_centers
        Camera center *[batch_shape,2,3,1]*
    world_ray_vectors
        World ray vectors *[batch_shape,2,image_shape,3]*
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image dimensions. Inferred from inputs in None. (Default value = None)
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        Closest mutual points image *[batch_shape,2,image_shape,4]*

    """
    if batch_shape is None:
        batch_shape = camera_centers.shape[:-3]
    num_batch_dims = len(batch_shape)

    if image_shape is None:
        image_shape = world_ray_vectors.shape[num_batch_dims + 1 : -1]
    num_image_dims = len(image_shape)

    if device is None:
        device = ivy.dev(camera_centers)

    # shapes as list
    batch_shape = list(batch_shape)
    image_shape = list(image_shape)

    # BS x 3 x 1
    camera_center0 = camera_centers[..., 0, :, :]
    camera_center1 = camera_centers[..., 1, :, :]

    # BS x [1]*NID x 3
    cam1_to_cam2 = ivy.reshape(
        camera_center1 - camera_center0, batch_shape + [1] * num_image_dims + [3]
    )
    cam2_to_cam1 = ivy.reshape(
        camera_center0 - camera_center1, batch_shape + [1] * num_image_dims + [3]
    )

    # BS x 2 x IS x 3
    ds = world_ray_vectors

    # BS x IS x 3
    ds0 = ds[tuple([slice(None)] * num_batch_dims + [0])]
    ds1 = ds[tuple([slice(None)] * num_batch_dims + [1])]
    n = ivy.cross(ds0, ds1)
    n1 = ivy.cross(ds0, n)
    n2 = ivy.cross(ds1, n)

    # BS x 1 x IS
    t1 = ivy.expand_dims(
        ivy.sum(cam1_to_cam2 * n2, axis=-1)
        / (ivy.sum(ds0 * n2, axis=-1) + MIN_DENOMINATOR),
        axis=num_batch_dims,
    )
    t2 = ivy.expand_dims(
        ivy.sum(cam2_to_cam1 * n1, axis=-1)
        / (ivy.sum(ds1 * n1, axis=-1) + MIN_DENOMINATOR),
        axis=num_batch_dims,
    )

    # BS x 2 x IS x 1
    ts = ivy.expand_dims(ivy.concat((t1, t2), axis=num_batch_dims), axis=-1)

    # BS x 2 x IS x 3
    world_coords = (
        ivy.reshape(
            camera_centers[..., 0], batch_shape + [2] + [1] * num_image_dims + [3]
        )
        + ts * world_ray_vectors
    )

    # BS x 2 x IS x 4
    return ivy_mech.make_coordinates_homogeneous(
        world_coords, batch_shape + [2] + image_shape
    )


def _triangulate_depth_by_closest_mutual_points(
    ds_pixel_coords, full_mats, inv_full_mats, camera_centers, batch_shape, image_shape
):
    """

    Parameters
    ----------
    ds_pixel_coords

    full_mats

    inv_full_mats

    camera_centers

    batch_shape

    image_shape

    """
    # single view geom batch shape
    svg_batch_shape = batch_shape + [2]
    num_batch_dims = len(batch_shape)

    # BS x 2 x IS x 3
    world_rays_flat = ivy_svg.pixel_coords_to_world_ray_vectors(
        inv_full_mats, ds_pixel_coords, camera_centers, svg_batch_shape, image_shape
    )

    # BS x 2 x IS x 3
    world_rays = ivy.reshape(world_rays_flat, svg_batch_shape + image_shape + [3])

    # BS x 2 x IS x 4
    world_points = closest_mutual_points_along_two_skew_rays(
        camera_centers, world_rays, batch_shape, image_shape
    )

    # BS x IS x 3
    return ivy_svg.world_to_ds_pixel_coords(
        world_points[tuple([slice(None)] * num_batch_dims + [0])],
        full_mats[..., 0, :, :],
        batch_shape,
        image_shape,
    )


def _triangulate_depth_by_homogeneous_dlt(
    ds_pixel_coords, full_mats, _, _1, batch_shape, image_shape
):
    """

    Parameters
    ----------
    ds_pixel_coords

    full_mats

    _

    _1

    batch_shape

    image_shape

    """
    # num batch dims
    num_batch_dims = len(batch_shape)
    num_image_dims = len(image_shape)
    image_size = reduce(mul, image_shape)

    # BS x 2 x IS x 3
    pixel_coords_normalized = ds_pixel_coords / (
        ds_pixel_coords[..., -1:] + MIN_DENOMINATOR
    )

    # BS x 3 x 4
    P = full_mats[..., 0, :, :]
    P_dash = full_mats[..., 1, :, :]

    # BS x (IS) x 4
    p1T = ivy.tile(P[..., 0:1, :], [1] * num_batch_dims + [image_size, 1])
    p2T = ivy.tile(P[..., 1:2, :], [1] * num_batch_dims + [image_size, 1])
    p3T = ivy.tile(P[..., 2:3, :], [1] * num_batch_dims + [image_size, 1])

    p_dash_1T = ivy.tile(P_dash[..., 0:1, :], [1] * num_batch_dims + [image_size, 1])
    p_dash_2T = ivy.tile(P_dash[..., 1:2, :], [1] * num_batch_dims + [image_size, 1])
    p_dash_3T = ivy.tile(P_dash[..., 2:3, :], [1] * num_batch_dims + [image_size, 1])

    # BS x (IS) x 1
    x = ivy.reshape(
        pixel_coords_normalized[
            tuple(
                [slice(None)] * num_batch_dims
                + [0]
                + [slice(None)] * num_image_dims
                + [0]
            )
        ],
        batch_shape + [image_size, 1],
    )
    y = ivy.reshape(
        pixel_coords_normalized[
            tuple(
                [slice(None)] * num_batch_dims
                + [0]
                + [slice(None)] * num_image_dims
                + [1]
            )
        ],
        batch_shape + [image_size, 1],
    )
    x_dash = ivy.reshape(
        pixel_coords_normalized[
            tuple(
                [slice(None)] * num_batch_dims
                + [1]
                + [slice(None)] * num_image_dims
                + [0]
            )
        ],
        batch_shape + [image_size, 1],
    )
    y_dash = ivy.reshape(
        pixel_coords_normalized[
            tuple(
                [slice(None)] * num_batch_dims
                + [1]
                + [slice(None)] * num_image_dims
                + [1]
            )
        ],
        batch_shape + [image_size, 1],
    )

    # BS x (IS) x 1 x 4
    A_row1 = ivy.expand_dims(x * p3T - p1T, axis=-2)
    A_row2 = ivy.expand_dims(y * p3T - p2T, axis=-2)
    A_row3 = ivy.expand_dims(x_dash * p_dash_3T - p_dash_1T, axis=-2)
    A_row4 = ivy.expand_dims(y_dash * p_dash_3T - p_dash_2T, axis=-2)

    # BS x (IS) x 4 x 4
    A = ivy.concat((A_row1, A_row2, A_row3, A_row4), axis=-2)

    # BS x (IS) x 4
    X = ivy_pg.solve_homogeneous_dlt(A)

    # BS x IS x 4
    coords_wrt_world_homo_unscaled = ivy.reshape(X, batch_shape + image_shape + [4])
    coords_wrt_world = coords_wrt_world_homo_unscaled / (
        coords_wrt_world_homo_unscaled[..., -1:] + MIN_DENOMINATOR
    )

    # BS x IS x 3
    return ivy_svg.world_to_ds_pixel_coords(
        coords_wrt_world, full_mats[..., 0, :, :], batch_shape, image_shape
    )


TRI_METHODS = {
    "cmp": _triangulate_depth_by_closest_mutual_points,
    "dlt": _triangulate_depth_by_homogeneous_dlt,
}


def triangulate_depth(
    ds_pixel_coords,
    full_mats,
    inv_full_mats=None,
    camera_centers=None,
    method="cmp",
    batch_shape=None,
    image_shape=None,
):
    r"""Triangulate depth in frame 1, returning depth scaled homogeneous pixel
    co-ordinate image :math:`\mathbf{X}\in\mathbb{R}^{is×3}` in frame 1.\n

    Parameters
    ----------
    ds_pixel_coords
        Homogeneous pixel co-ordinate images: *[batch_shape,image_shape,3]*
    full_mats
        Full projection matrices *[batch_shape,2,3,4]*
    inv_full_mats
        Inverse full projection matrices, required for closest_mutual_points method
        *[batch_shape,2,3,4]* (Default value = None)
    camera_centers
        Camera centers, required for closest_mutual_points method *[batch_shape,2,3,1]*
        (Default value = None)
    method
        Triangulation method, one of [cmp|dlt], for closest mutual points or
        homogeneous dlt approach, closest_mutual_points by default
    batch_shape
        Shape of batch. Inferred from inputs if None. (Default value = None)
    image_shape
        Image dimensions. Inferred from inputs in None. (Default value = None)

    Returns
    -------
    ret
        Depth scaled homogeneous pixel co-ordinates image in frame 1
        *[batch_shape,image_shape,3]*

    """
    if batch_shape is None:
        batch_shape = ds_pixel_coords.shape[:-4]

    if image_shape is None:
        image_shape = ds_pixel_coords.shape[-3:-1]

    # shapes as list
    batch_shape = list(batch_shape)
    image_shape = list(image_shape)

    if method == "cmt":
        if inv_full_mats is None:
            inv_full_mats = ivy.inv(
                ivy_mech.make_transformation_homogeneous(full_mats, batch_shape + [2])
            )[..., 0:3, :]

        if camera_centers is None:
            camera_centers = ivy_svg.inv_ext_mat_to_camera_center(inv_full_mats)

    try:
        return TRI_METHODS[method](
            ds_pixel_coords,
            full_mats,
            inv_full_mats,
            camera_centers,
            batch_shape,
            image_shape,
        )
    except KeyError:
        raise Exception(
            "Triangulation method must be one of [cmp|dlt], but found {}".format(method)
        )



# global
import ivy

# local
import ivy_vision.sdf as ivy_sdf


# noinspection PyMissingConstructor
class PrimitiveScene(ivy.Container):
    def __init__(
        self,
        sphere_positions=None,
        sphere_radii=None,
        cuboid_ext_mats=None,
        cuboid_dims=None,
    ):
        """Initialize scene description as a composition of primitive shapes. # ToDo:
        extend this to include cylinders and cones once supported in ivy_vision.sdf
        module

        Parameters
        ----------
        sphere_positions
            Sphere positions *[batch_shape,num_spheres,3]* (Default value = None)
        sphere_radii
            Sphere radii *[batch_shape,num_spheres,1]* (Default value = None)
        cuboid_ext_mats
            Cuboid inverse extrinsic matrices *[batch_shape,num_cuboids,3,4]*
            (Default value = None)
        cuboid_dims
            Cuboid dimensions, in order of x, y, z *[batch_shape,num_cuboids,3]*
            (Default value = None)

        """
        super(PrimitiveScene, self).__init__(
            sphere_positions=sphere_positions,
            sphere_radii=sphere_radii,
            cuboid_ext_mats=cuboid_ext_mats,
            cuboid_dims=cuboid_dims,
        )

    # Class Methods #
    # --------------#

    @staticmethod
    def as_identity(batch_shape):
        """Return primitive scene object with array attributes as either zeros or
        identity matrices.

        Parameters
        ----------
        batch_shape
            Batch shape for each geometric array attribute

        Returns
        -------
        ret
            New primitive scene object, with each entry as either zeros or
            identity matrices.

        """
        batch_shape = list(batch_shape)
        sphere_positions = ivy.eye(4, batch_shape=batch_shape)[..., 0:3, :]
        sphere_radii = ivy.ones(batch_shape + [1])
        cuboid_ext_mats = ivy.eye(4, batch_shape=batch_shape)[..., 0:3, :]
        cuboid_dims = ivy.ones(batch_shape + [3])
        return __class__(sphere_positions, sphere_radii, cuboid_ext_mats, cuboid_dims)

    # Public Methods #
    # ---------------#

    def sdf(self, query_positions):
        """Return signed distance function for the scene

        Parameters
        ----------
        query_positions
            Point for which to query the signed distance *[batch_shape,num_points,3]*

        Returns
        -------
        ret
            The signed distance values for each of the query points in the scene
            *[batch_shape,num_points,1]*

        """
        # BS x NP x 1
        all_sdfs_list = list()
        if self.sphere_positions is not None:
            sphere_sdfs = ivy_sdf.sphere_signed_distances(
                self.sphere_positions[..., 0:3, -1], self.sphere_radii, query_positions
            )
            all_sdfs_list.append(sphere_sdfs)
        if self.cuboid_ext_mats is not None:
            cuboid_sdfs = ivy_sdf.cuboid_signed_distances(
                self.cuboid_ext_mats, self.cuboid_dims, query_positions
            )
            all_sdfs_list.append(cuboid_sdfs)
        sdfs_concatted = (
            ivy.concat(all_sdfs_list, axis=-1)
            if len(all_sdfs_list) > 1
            else all_sdfs_list[0]
        )
        return ivy.min(sdfs_concatted, axis=-1, keepdims=True)


# noinspection PyMissingConstructor
class Intrinsics(ivy.Container):
    def __init__(
        self, focal_lengths, persp_angles, pp_offsets, calib_mats, inv_calib_mats
    ):
        """Initialize camera intrinsics container.

        Parameters
        ----------
        focal_lengths
            Focal lengths *[batch_shape,2]*
        persp_angles
            perspective angles *[batch_shape,2]*
        pp_offsets
            Principal-point offsets *[batch_shape,2]*
        calib_mats
            Calibration matrices *[batch_shape,3,3]*
        inv_calib_mats
            Inverse calibration matrices *[batch_shape,3,3]*

        """
        super(Intrinsics, self).__init__(
            focal_lengths=focal_lengths,
            persp_angles=persp_angles,
            pp_offsets=pp_offsets,
            calib_mats=calib_mats,
            inv_calib_mats=inv_calib_mats,
        )

    # Class Methods #
    # --------------#

    @staticmethod
    def as_identity(batch_shape):
        """Return camera intrinsics object with array attributes as either zeros or
        identity matrices.

        Parameters
        ----------
        batch_shape
            Batch shape for each geometric array attribute

        Returns
        -------
        ret
            New camera intrinsics object, with each entry as either zeros or
            identity matrices.

        """
        batch_shape = list(batch_shape)
        focal_lengths = ivy.ones(batch_shape + [2])
        persp_angles = ivy.ones(batch_shape + [2])
        pp_offsets = ivy.zeros(batch_shape + [2])
        calib_mats = ivy.eye(3, batch_shape=batch_shape)
        inv_calib_mats = ivy.eye(3, batch_shape=batch_shape)
        return __class__(
            focal_lengths, persp_angles, pp_offsets, calib_mats, inv_calib_mats
        )


# noinspection PyMissingConstructor
class Extrinsics(ivy.Container):
    def __init__(self, cam_centers, Rs, inv_Rs, ext_mats_homo, inv_ext_mats_homo):
        """Initialize camera extrinsics container.

        Parameters
        ----------
        cam_centers
            Camera centers *[batch_shape,3,1]*
        Rs
            Rotation matrices *[batch_shape,3,3]*
        inv_Rs
            Inverse rotation matrices *[batch_shape,3,3]*
        ext_mats_homo
            Homogeneous extrinsic matrices *[batch_shape,4,4]*
        inv_ext_mats_homo
            Inverse homogeneous extrinsic matrices *[batch_shape,4,4]*

        """
        super(Extrinsics, self).__init__(
            cam_centers=cam_centers,
            Rs=Rs,
            inv_Rs=inv_Rs,
            ext_mats_homo=ext_mats_homo,
            inv_ext_mats_homo=inv_ext_mats_homo,
        )

    # Class Methods #
    # --------------#

    @staticmethod
    def as_identity(batch_shape):
        """Return camera extrinsics object with array attributes as either zeros or
        identity matrices.

        Parameters
        ----------
        batch_shape
            Batch shape for each geometric array attribute.

        Returns
        -------
        ret
            New camera extrinsics object, with each entry as either zeros or
            identity matrices.

        """
        batch_shape = list(batch_shape)
        cam_centers = ivy.zeros(batch_shape + [3, 1])
        Rs = ivy.eye(3, batch_shape=batch_shape)
        inv_Rs = ivy.eye(3, batch_shape=batch_shape)
        ext_mats_homo = ivy.eye(4, batch_shape=batch_shape)
        inv_ext_mats_homo = ivy.eye(4, batch_shape=batch_shape)
        return __class__(cam_centers, Rs, inv_Rs, ext_mats_homo, inv_ext_mats_homo)


# noinspection PyMissingConstructor
class CameraGeometry(ivy.Container):
    def __init__(self, intrinsics, extrinsics, full_mats_homo, inv_full_mats_homo):
        """Initialize camera geometry container.

        Parameters
        ----------
        intrinsics
            Camera intrinsics object.
        extrinsics
            Camera extrinsics object.
        full_mats_homo
            Full homogeneous projection matrices *[batch_shape,4,4]*
        inv_full_mats_homo
            Inverse full homogeneous projection matrices *[batch_shape,4,4]*

        """
        super(CameraGeometry, self).__init__(
            intrinsics=intrinsics,
            extrinsics=extrinsics,
            full_mats_homo=full_mats_homo,
            inv_full_mats_homo=inv_full_mats_homo,
        )

    # Class Methods #
    # --------------#

    @staticmethod
    def as_identity(batch_shape):
        """Return camera geometry object with array attributes as either zeros or
        identity matrices.

        Parameters
        ----------
        batch_shape
            Batch shape for each geometric array attribute

        Returns
        -------
        ret
            New camera geometry object, with each entry as either zeros or
            identity matrices.

        """
        intrinsics = Intrinsics.as_identity(batch_shape)
        extrinsics = Extrinsics.as_identity(batch_shape)
        full_mats_homo = ivy.eye(4, batch_shape=batch_shape)
        inv_full_mats_homo = ivy.eye(4, batch_shape=batch_shape)
        return __class__(intrinsics, extrinsics, full_mats_homo, inv_full_mats_homo)



"""Collection of Smoothing Functions"""

# global
import ivy
import math

# local
from ivy_vision import single_view_geometry as ivy_svg

MIN_DENOMINATOR = 1e-12


# noinspection PyUnresolvedReferences
def weighted_image_smooth(mean, weights, kernel_dim):
    """Smooth an image using weight values from a weight image of the same size.

    Parameters
    ----------
    mean
        Image to smooth *[batch_shape,h,w,d]*
    weights
        Variance image, with the variance values of each pixel in the image
        *[batch_shape,h,w,d]*
    kernel_dim
        The dimension of the kernel

    Returns
    -------
    ret
        Image smoothed based on variance image and smoothing kernel.

    """
    # shapes as list
    kernel_shape = [kernel_dim, kernel_dim]
    dim = mean.shape[-1]

    # KW x KW x D
    kernel = ivy.ones(kernel_shape + [dim])

    # D
    kernel_sum = ivy.sum(kernel, axis=[0, 1])[0]

    # BS x H x W x D
    mean_x_weights = mean * weights
    mean_x_weights_sum = ivy.abs(
        ivy.depthwise_conv2d(
            mean_x_weights.astype("float64"), kernel.astype("float64"), 1, "VALID"
        )
    )
    sum_of_weights = ivy.depthwise_conv2d(
        weights.astype("float64"), kernel.astype("float64"), 1, "VALID"
    )
    new_mean = mean_x_weights_sum / (sum_of_weights + MIN_DENOMINATOR)

    new_weights = sum_of_weights / (kernel_sum + MIN_DENOMINATOR)

    # BS x H x W x D,  # BS x H x W x D
    return new_mean, new_weights


def smooth_image_fom_var_image(mean, var, kernel_dim, kernel_scale, device=None):
    """Smooth an image using variance values from a variance image of the same size,
    and a spatial smoothing kernel.

    Parameters
    ----------
    mean
        Image to smooth *[batch_shape,h,w,d]*
    var
        Variance image, with the variance values of each pixel in the image
        *[batch_shape,h,w,d]*
    kernel_dim
        The dimension of the kernel
    kernel_scale
        The scale of the kernel along the channel dimension *[d]*
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as x if None. (Default value = None)

    Returns
    -------
    ret
        Image smoothed based on variance image and smoothing kernel.

    """
    if device is None:
        device = ivy.dev(mean)

    # shapes as list
    kernel_shape = [kernel_dim, kernel_dim]
    kernel_size = kernel_dim**2
    dims = mean.shape[-1]

    # KH x KW x 2
    uniform_pixel_coords = ivy_svg.create_uniform_pixel_coords_image(
        kernel_shape, device=device
    )[..., 0:2]

    # 2
    kernel_central_pixel_coord = ivy.array(
        [
            float(math.floor(kernel_shape[0] / 2)),
            float(math.floor(kernel_shape[1] / 2)),
        ],
        device=device,
    )

    # KH x KW x 2
    kernel_xy_dists = kernel_central_pixel_coord - uniform_pixel_coords
    kernel_xy_dists_sqrd = kernel_xy_dists**2

    # KW x KW x D x D
    unit_kernel = ivy.tile(
        ivy.sum(kernel_xy_dists_sqrd, axis=-1, keepdims=True) ** 0.5, (1, 1, dims)
    )
    kernel = 1 + unit_kernel * kernel_scale
    recip_kernel = 1 / (kernel + MIN_DENOMINATOR)

    # D
    kernel_sum = ivy.sum(kernel, axis=[0, 1])[0]
    recip_kernel_sum = ivy.sum(recip_kernel, axis=[0, 1])

    # BS x H x W x D
    recip_var = 1 / (var + MIN_DENOMINATOR)
    recip_var_scaled = recip_var + 1

    recip_new_var_scaled = ivy.depthwise_conv2d(
        recip_var_scaled, recip_kernel, 1, "VALID"
    )
    # This 0.99 prevents float32 rounding errors leading to -ve variances,
    # the true equation would use 1.0
    recip_new_var = recip_new_var_scaled - recip_kernel_sum * 0.99
    new_var = 1 / (recip_new_var + MIN_DENOMINATOR)

    mean_x_recip_var = mean * recip_var
    mean_x_recip_var_sum = ivy.abs(
        ivy.depthwise_conv2d(mean_x_recip_var, recip_kernel, 1, "VALID")
    )
    new_mean = new_var * mean_x_recip_var_sum

    new_var = new_var * kernel_size**2 / (kernel_sum + MIN_DENOMINATOR)
    # prevent overconfidence from false meas independence assumption

    # BS x H x W x D,        # BS x H x W x D
    return new_mean, new_var



"""Collection of Singed Distance Functions"""
# ToDo: extend to include cylinders and cones
# https://www.iquilezles.org/www/articles/distfunctions/distfunctions.htm

# global
import ivy


def sphere_signed_distances(sphere_positions, sphere_radii, query_positions):
    r"""Return the signed distances of a set of query points from the sphere
    surfaces.\n `[reference] <https://www.iquilezles.org/www/articles/distfunctions
    /distfunctions.htm>`_

    Parameters
    ----------
    sphere_positions
        Positions of the spheres *[batch_shape,num_spheres,3]*
    sphere_radii
        Radii of the spheres *[batch_shape,num_spheres,1]*
    query_positions
        Points for which to query the signed distances *[batch_shape,num_points,3]*

    Returns
    -------
    ret
        The distances of the query points from the closest sphere surface
        *[batch_shape,num_points,1]*

    """
    # BS x NS x 1 x 3
    sphere_positions = ivy.expand_dims(sphere_positions, axis=-2)

    # BS x 1 x NP x 3
    query_positions = ivy.expand_dims(query_positions, axis=-3)

    # BS x NS x NP x 1
    distances_to_centre = (
        ivy.sum((query_positions - sphere_positions) ** 2, axis=-1, keepdims=True)
        ** 0.5
    )

    # BS x NS x NP x 1
    all_sdfs = distances_to_centre - ivy.expand_dims(sphere_radii, axis=-2)

    # BS x NP x 1
    return ivy.min(all_sdfs, axis=-3)


def cuboid_signed_distances(
    cuboid_ext_mats, cuboid_dims, query_positions, batch_shape=None
):
    r"""Return the signed distances of a set of query points from the cuboid
    surfaces.\n `[reference] <https://www.iquilezles.org/www/articles/distfunctions
    /distfunctions.htm>`_

    Parameters
    ----------
    cuboid_ext_mats
        Extrinsic matrices of the cuboids *[batch_shape,num_cuboids,3,4]*
    cuboid_dims
        Dimensions of the cuboids, in the order x, y, z *[batch_shape,num_cuboids,3]*
    query_positions
        Points for which to query the signed distances *[batch_shape,num_points,3]*
    batch_shape
        Shape of batch. Assumed no batches if None. (Default value = None)

    Returns
    -------
    ret
        The distances of the query points from the closest cuboid surface
        *[batch_shape,num_points,1]*

    """
    if batch_shape is None:
        batch_shape = cuboid_ext_mats.shape[:-3]

    # shapes as list
    batch_shape = list(batch_shape)
    num_batch_dims = len(batch_shape)
    batch_dims_for_trans = list(range(num_batch_dims))
    num_cuboids = cuboid_ext_mats.shape[-3]
    num_points = query_positions.shape[-2]

    # BS x 3 x NP
    query_positions_trans = ivy.permute_dims(
        query_positions,
        axes=batch_dims_for_trans + [num_batch_dims + 1, num_batch_dims],
    )

    # BS x 1 x NP
    ones = ivy.ones_like(query_positions_trans[..., 0:1, :])

    # BS x 4 x NP
    query_positions_trans_homo = ivy.concat((query_positions_trans, ones), axis=-2)

    # BS x NCx3 x 4
    cuboid_ext_mats_flat = ivy.reshape(cuboid_ext_mats, batch_shape + [-1, 4])

    # BS x NCx3 x NP
    rel_query_positions_trans_flat = ivy.matmul(
        cuboid_ext_mats_flat, query_positions_trans_homo
    )

    # BS x NC x 3 x NP
    rel_query_positions_trans = ivy.reshape(
        rel_query_positions_trans_flat, batch_shape + [num_cuboids, 3, num_points]
    )

    # BS x NC x NP x 3
    rel_query_positions = ivy.permute_dims(
        rel_query_positions_trans,
        axes=batch_dims_for_trans
        + [num_batch_dims, num_batch_dims + 2, num_batch_dims + 1],
    )
    q = ivy.abs(rel_query_positions) - ivy.expand_dims(cuboid_dims / 2, axis=-2)
    q_max_clipped = ivy.maximum(q, 1e-12)

    # BS x NC x NP x 1
    q_min_clipped = ivy.minimum(ivy.max(q, axis=-1, keepdims=True), 0.0)
    q_max_clipped_len = ivy.sum(q_max_clipped**2, axis=-1, keepdims=True) ** 0.5
    sdfs = q_max_clipped_len + q_min_clipped

    # BS x NP x 1
    return ivy.min(sdfs, axis=-3)



# global
import ivy
import math
import ivy_mech
import numpy as np
from operator import mul
from functools import reduce

# local
from ivy_vision import single_view_geometry as ivy_svg

MIN_DENOMINATOR = 1e-12


def downsampled_image_dims_from_desired_num_pixels(
    image_dims, num_pixels, maximum=False
):
    """Compute the best downsampled image dimensions, given original image dimensions
    and the ideal total number of pixels for the downsampled image. The final number
    of pixels in the downsampled image dimensions may not be exact, but will best
    balance aspect ratio preservation and total number of pixels.

    Parameters
    ----------
    image_dims
        The image dimensions of the original image.
    num_pixels
        The ideal numbr of pixels in the downsampled image.
    maximum
        Whether the number of pixels is a hard maximum. Default is False.

    Returns
    -------
    ret
        The image dimensions for a new downsampled image.

    """
    int_method = math.floor if maximum else round
    ratio = image_dims[1] / image_dims[0]
    new_dim_0 = (num_pixels / ratio) ** 0.5
    new_dim_1 = int_method(new_dim_0 * ratio)
    new_dim_0 = int_method(new_dim_0)
    new_img_dims = [new_dim_0, new_dim_1]
    num_pixels = new_dim_0 * new_dim_1
    return new_img_dims, num_pixels


def create_sampled_pixel_coords_image(
    image_dims,
    samples_per_dim,
    batch_shape=None,
    normalized=False,
    randomize=True,
    homogeneous=False,
    device=None,
):
    r"""Create image of randomly sampled homogeneous integer :math:`xy` pixel
    co-ordinates :math:`\mathbf{X}\in\mathbb{Z}^{h×w×3}`, stored as floating point
    values. The origin is at the top-left corner of the image, with :math:`+x`
    rightwards, and :math:`+y` downwards. The final homogeneous dimension are all
    ones. In subsequent use of this image, the depth of each pixel can be represented
    using this same homogeneous representation, by simply scaling each 3-vector by
    the depth value. The final dimension therefore always holds the depth value,
    while the former two dimensions hold depth scaled pixel co-ordinates.\n `[
    reference] <localhost:63342/ivy/docs/source/references/mvg_textbook.pdf#page=172
    >`_ deduction from top of page 154, section 6.1, equation 6.1

    Parameters
    ----------
    image_dims
        Image dimensions.
    samples_per_dim
        The number of samples to perform per image dimension.
    batch_shape
        Shape of batch. Assumed no batch dimensions if None. (Default value = None)
    normalized
        Whether to normalize x-y pixel co-ordinates to the range 0-1. Default is False.
    randomize
        Whether to randomize the sampled co-ordiantes within their window.
        Default is True.
    homogeneous
        Whether the pixel co-ordinates should be 3D homogeneous or just 2D.
        Default is True.
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        (Default value = None)

    Returns
    -------
    ret
        Image of homogeneous pixel co-ordinates *[batch_shape,height,width,3]*

    """
    # BS x DH x DW x 2
    low_res_pix_coords = ivy_svg.create_uniform_pixel_coords_image(
        samples_per_dim, batch_shape, homogeneous=False, device=device
    )

    # 2
    window_size = ivy.array(
        list(
            reversed(
                [
                    img_dim / sam_per_dim
                    for img_dim, sam_per_dim in zip(image_dims, samples_per_dim)
                ]
            )
        ),
        device=device,
    )

    # BS x DH x DW x 2
    downsam_pix_coords = low_res_pix_coords * window_size + window_size / 2 - 0.5

    if randomize:
        # BS x DH x DW x 1
        rand_x = ivy.random_uniform(
            low=ivy.to_scalar(-window_size[0] / 2),
            high=ivy.to_scalar(window_size[0] / 2),
            shape=list(downsam_pix_coords.shape[:-1]) + [1],
            device=device,
        )
        rand_y = ivy.random_uniform(
            low=ivy.to_scalar(-window_size[1] / 2),
            high=ivy.to_scalar(window_size[1] / 2),
            shape=list(downsam_pix_coords.shape[:-1]) + [1],
            device=device,
        )

        # BS x DH x DW x 2
        rand_offsets = ivy.concat((rand_x, rand_y), axis=-1)
        downsam_pix_coords += rand_offsets
    downsam_pix_coords = ivy.clip(
        ivy.round(downsam_pix_coords),
        ivy.array([0.0] * 2, device=device),
        ivy.array(list(reversed(image_dims)), dtype="float32", device=device) - 1,
    )

    if normalized:
        downsam_pix_coords /= (
            ivy.array([image_dims[1], image_dims[0]], dtype="float32", device=device)
            + MIN_DENOMINATOR
        )

    if homogeneous:
        # BS x DH x DW x 3
        downsam_pix_coords = ivy_mech.make_coordinates_homogeneous(
            downsam_pix_coords, batch_shape + samples_per_dim
        )

    # BS x DH x DW x 2or3
    return downsam_pix_coords


def sample_images(list_of_images, num_pixels, batch_shape, image_dims, device=None):
    """Samples each image in a list of aligned images at num_pixels random pixel
    co-ordinates, within a unifrom grid over the image.

    Parameters
    ----------
    list_of_images
        List of images to sample from.
    num_pixels
        The number of pixels to sample from each image.
    batch_shape
        Shape of batch. Inferred from inputs if None.
    image_dims
        Image dimensions. Inferred from inputs in None.
    device
        device on which to create the array 'cuda:0', 'cuda:1', 'cpu' etc.
        Same as images if None. (Default value = None)

    """
    if batch_shape is None:
        batch_shape = list_of_images[0].shape[:-3]

    if image_dims is None:
        image_dims = list_of_images[0].shape[-3:-1]

    if device is None:
        device = ivy.dev(list_of_images[0])

    image_channels = [img.shape[-1] for img in list_of_images]

    # shapes as list
    batch_shape = list(batch_shape)
    image_dims = list(image_dims)
    flat_batch_size = reduce(mul, batch_shape)

    new_img_dims, num_pixels_to_use = downsampled_image_dims_from_desired_num_pixels(
        image_dims, num_pixels
    )

    # BS x DH x DW x 2
    sampled_pix_coords = ivy.astype(
        create_sampled_pixel_coords_image(
            image_dims, new_img_dims, batch_shape, homogeneous=False, device=device
        ),
        "int32",
    )

    # FBS x DH x DW x 2
    sampled_pix_coords_flat = ivy.reshape(
        sampled_pix_coords, [flat_batch_size] + new_img_dims + [2]
    )

    # FBS x DH x DW x 1
    batch_idxs = ivy.expand_dims(
        ivy.permute_dims(
            ivy.astype(
                ivy.linspace(
                    ivy.zeros(new_img_dims, device=device),
                    ivy.ones(new_img_dims, device=device) * (flat_batch_size - 1),
                    flat_batch_size,
                    axis=-1,
                ),
                "int32",
            ),
            axes=(2, 0, 1),
        ),
        axis=-1,
    )

    # FBS x DH x DW x 3
    total_idxs = ivy.concat(
        (batch_idxs, ivy.flip(sampled_pix_coords_flat, axis=-1)), axis=-1
    )

    # list of FBS x H x W x D
    flat_batch_images = [
        ivy.reshape(img, [flat_batch_size] + image_dims + [-1])
        for img in list_of_images
    ]

    # FBS x H x W x sum(D)
    combined_img = ivy.concat(flat_batch_images, axis=-1)

    # BS x FID x sum(D)
    combined_img_sampled = ivy.reshape(
        ivy.gather_nd(combined_img, total_idxs), batch_shape + [num_pixels_to_use, -1]
    )

    # list of BS x FID x D
    return ivy.split(combined_img_sampled, num_or_size_splits=image_channels, axis=-1)


# noinspection PyUnresolvedReferences
def sampled_volume_density_to_occupancy_probability(density, inter_sample_distance):
    """Compute probability of occupancy, given sampled volume densities and their
    associated inter-sample distances

    Parameters
    ----------
    density
        The sampled density values *[batch_shape]*
    inter_sample_distance
        The inter-sample distances *[batch_shape]*

    Returns
    -------
    ret
        The occupancy probabilities *[batch_shape]*

    """
    return 1 - ivy.exp(-density * inter_sample_distance)


def ray_termination_probabilities(density, inter_sample_distance):
    """Compute probability of occupancy, given sampled volume densities and their
    associated inter-sample distances

    Parameters
    ----------
    density
        The sampled density values *[batch_shape,num_samples_per_ray]*
    inter_sample_distance
        The inter-sample distances *[batch_shape,num_samples_per_ray]*

    Returns
    -------
    ret
        The occupancy probabilities *[batch_shape]*

    """
    # BS x NSPR
    occ_prob = sampled_volume_density_to_occupancy_probability(
        density, inter_sample_distance
    )
    return occ_prob * ivy.cumprod(1.0 - occ_prob + 1e-10, axis=-1, exclusive=True)


# noinspection PyUnresolvedReferences
def stratified_sample(starts, ends, num_samples, batch_shape=None):
    """Perform stratified sampling, between start and end arrays. This operation
    divides the range into equidistant bins, and uniformly samples value within the
    ranges for each of these bins.

    Parameters
    ----------
    starts
        Start values *[batch_shape]*
    ends
        End values *[batch_shape]*
    num_samples
        The number of samples to generate between starts and ends
    batch_shape
        Shape of batch, Inferred from inputs if None. (Default value = None)

    Returns
    -------
    ret
        The stratified samples, with each randomly placed in uniformly spaced bins
        *[batch_shape,num_samples]*

    """
    # shapes
    if batch_shape is None:
        batch_shape = starts.shape

    # shapes as lists
    batch_shape = list(batch_shape)

    # BS
    bin_sizes = (ends - starts) / num_samples

    # BS x NS
    linspace_vals = ivy.linspace(starts, ends - bin_sizes, num_samples)

    # BS x NS
    random_uniform = ivy.random_uniform(
        shape=batch_shape + [num_samples], device=ivy.dev(starts)
    )

    # BS x NS
    random_offsets = random_uniform * ivy.expand_dims(bin_sizes, axis=-1)

    # BS x NS
    return linspace_vals + random_offsets


# noinspection PyUnresolvedReferences
def render_rays_via_termination_probabilities(
    ray_term_probs, features, render_variance=False
):
    """Render features onto the image plane, given rays sampled at radial depths with
    readings of feature values and densities at these sample points.

    Parameters
    ----------
    ray_term_probs
        The ray termination probabilities *[batch_shape,num_samples_per_ray]*
    features
        Feature values at the sample points *[batch_shape,num_samples_per_ray,feat_dim]*
    render_variance
        Whether to also render the feature variance. Default is False.

    Returns
    -------
    ret
        The feature renderings along the rays, computed via the termination
        probabilities *[batch_shape,feat_dim]*

    """
    # BS x NSPR
    rendering = ivy.sum(ivy.expand_dims(ray_term_probs, axis=-1) * features, axis=-2)
    if not render_variance:
        return rendering
    var = ivy.sum(
        ray_term_probs * (ivy.expand_dims(rendering, axis=-2) - features) ** 2, axis=-2
    )
    return rendering, var


def render_implicit_features_and_depth(
    network_fn,
    rays_o,
    rays_d,
    near,
    far,
    samples_per_ray,
    timestamps=None,
    render_depth=True,
    render_feats=True,
    render_variance=False,
    inter_feat_fn=None,
    v=None,
):
    """Render an rgb-d image, given an implicit rgb and density function conditioned
    on xyz data.

    Parameters
    ----------
    network_fn
        the implicit function.
    rays_o
        The camera center *[batch_shape,3]*
    rays_d
        The rays in world space *[batch_shape,ray_batch_shape,3]*
    near
        The near clipping plane values *[batch_shape,ray_batch_shape]*
    far
        The far clipping plane values *[batch_shape,ray_batch_shape]*
    samples_per_ray
        The number of stratified samples to use along each ray
    timestamps
        The timestamps associated with each image. Default is None. *[batch_shape,1]*
    render_depth
        Whether to render the depth. Default is True.
    render_feats
        Whether to render the features. Default is True.
    render_variance
        Whether to also render the feature variance. Default is False.
    inter_feat_fn
        Function to extract interpolated features from world-coords
        *[batch_shape,ray_batch_shape,3]* (Default value = None)
    v
        The container of trainable variables for the implicit model.
        Default is to use internal variables.

    Returns
    -------
    ret
        The rendered feature *[batch_shape,ray_batch_shape,feat]* and radial depth
        *[batch_shape,ray_batch_shape,1]*

    """
    # shapes
    batch_shape = list(rays_o.shape[:-1])
    num_batch_dims = len(batch_shape)
    ray_batch_shape = list(rays_d.shape[num_batch_dims:-1])
    flat_ray_batch_size = int(np.prod(ray_batch_shape))
    num_ray_batch_dims = len(ray_batch_shape)
    total_batch_shape = batch_shape + ray_batch_shape

    # Compute 3D query points

    # BS x RBS x SPR x 1
    z_vals = ivy.expand_dims(stratified_sample(near, far, samples_per_ray), axis=-1)

    # BS x RBS x 1 x 3
    rays_d = ivy.expand_dims(rays_d, axis=-2)
    rays_o = ivy.broadcast_to(
        ivy.reshape(rays_o, batch_shape + [1] * (num_ray_batch_dims + 1) + [3]),
        rays_d.shape,
    )

    # BS x RBS x SPR x 3
    pts = rays_o + rays_d * z_vals

    # flatten

    # BS x FRBS x SPR x 3
    pts_flat = ivy.reshape(pts, batch_shape + [flat_ray_batch_size, samples_per_ray, 3])

    # input features

    if inter_feat_fn is not None:
        # BS x RBS x SPR x F
        features = inter_feat_fn(pts)

        # BS x FRBS x SPR x F
        features_flat = ivy.reshape(
            features, batch_shape + [flat_ray_batch_size, samples_per_ray, -1]
        )

    else:
        features_flat = None

    # Run network

    # BSPQ x OF,    BSPQ
    feat, densities = network_fn(pts_flat, features_flat, timestamps, v=v)

    # BS x RBS x SPR
    densities = ivy.reshape(densities, total_batch_shape + [samples_per_ray])

    # BS x RBS x (SPR+1)
    z_vals_w_terminal = ivy.concat(
        (z_vals[..., 0], ivy.ones_like(z_vals[..., -1:, 0]) * 1e10), axis=-1
    )

    # BS x RBS x SPR
    depth_diffs = z_vals_w_terminal[..., 1:] - z_vals_w_terminal[..., :-1]
    ray_term_probs = ray_termination_probabilities(densities, depth_diffs)

    # return values
    ret_vals = list()

    if render_feats:
        # BS x RBS x SPR x OF
        feat = ivy.reshape(feat, total_batch_shape + [samples_per_ray, -1])

        # BS x RBS x OF
        feat = ivy.clip(
            render_rays_via_termination_probabilities(
                ray_term_probs, feat, render_variance
            ),
            0.0,
            1.0,
        )
        if render_variance:
            ret_vals += feat
        else:
            ret_vals.append(feat)

    if render_depth:
        # BS x RBS x 1
        radial_depth = render_rays_via_termination_probabilities(
            ray_term_probs, z_vals, render_variance
        )
        if render_variance:
            ret_vals += radial_depth
        else:
            ret_vals.append(radial_depth)

    # up to BS x RBS x OF, BS x RBS x OF, BS x RBS x 1, BS x RBS x 1
    return tuple(ret_vals)
